# ***************************************************************** #
# (C) Copyright IBM Corporation 2024.                               #
#                                                                   #
# The source code for this program is not published or otherwise    #
# divested of its trade secrets, irrespective of what has been      #
# deposited with the U.S. Copyright Office.                         #
# ***************************************************************** #
# SPDX-License-Identifier: Apache-2.0
"""This script is used to add kubernetes_lite specific post-processing to the
result of gopy. This includes renaming platform specific files, updating CFLAG/LDFLAGS
and deleting extra unneeded files.
"""

from pathlib import Path

from scripts.utils import SYSTEM, SystemTypes

import typer

app = typer.Typer()

CGO_CFLAGS = "#cgo CFLAGS: -Wno-error -Wno-implicit-function-declaration -Wno-int-conversion -Ofast"

CGO_LDFLAGS: str | None = None
if SYSTEM in {SystemTypes.DARWIN, SystemTypes.LINUX}:
    CGO_LDFLAGS = "#cgo LDFLAGS: -ldl"


@app.command()
def post_gen_cleanup(parent_dir: Path, wrapper_subpath: str):
    """Cleanup and update the generated wrapper

    Args:
        parent_dir (Path): The root directory of the library
        wrapper_subpath (str): The name of the wrapper directory
    """
    # Delete generated license/library files
    for del_file_name in ("LICENSE", "Makefile", "MANIFEST.in", "README.md", "setup.py"):
        del_file_path = parent_dir / del_file_name
        del_file_path.unlink(missing_ok=True)

    wrapper_dir = parent_dir / wrapper_subpath
    # Move platform specific wrapper file to generic
    if len(list(wrapper_dir.glob("_wrapper.*.h"))) > 0:
        wrapper_h_file = next(wrapper_dir.glob("_wrapper.*.h"))
        wrapper_h_file.rename(wrapper_dir / "_wrapper.h")

    # Update the CGO flags
    go_file = wrapper_dir / "wrapper.go"
    go_output_text = ""
    for line in go_file.read_text().split("\n"):
        if line.startswith("#cgo CFLAGS"):
            go_output_text += CGO_CFLAGS + "\n"
        elif CGO_LDFLAGS and line.startswith("#cgo LDFLAGS"):
            go_output_text += CGO_LDFLAGS + "\n"
        else:
            go_output_text += line + "\n"
    go_file.write_text(go_output_text)


if __name__ == "__main__":
    app()



# ***************************************************************** #
# (C) Copyright IBM Corporation 2024.                               #
#                                                                   #
# The source code for this program is not published or otherwise    #
# divested of its trade secrets, irrespective of what has been      #
# deposited with the U.S. Copyright Office.                         #
# ***************************************************************** #
# SPDX-License-Identifier: Apache-2.0
"""This module contains helper functions to serialize python objects into
go Slice_bytes and back
"""

from io import BytesIO
from typing import Any

from kubernetes_lite.wrapper._wrapper import Slice_byte_from_bytes
from kubernetes_lite.wrapper.go import Slice_byte

import orjson


def serialize(obj: Any) -> bytes:
    """Serialize a python object using orjson into bytes

    Args:
        obj (Any): The object to serialize

    Returns:
        bytes: The return bytes data
    """
    if isinstance(obj, bytes):
        return obj
    return orjson.dumps(obj)


def serialize_to_go(obj: Any) -> Slice_byte:
    """Serialize a python object into a go object

    Args:
        obj (Any): The object to serialize

    Returns:
        Slice_byte: Pointer to the go object
    """
    if isinstance(obj, Slice_byte):
        return obj
    if isinstance(obj, BytesIO):
        obj = obj.read()
    # Don't use Slice_byte(<>) since its slow!
    handle = Slice_byte_from_bytes(serialize(obj))
    return Slice_byte(handle=handle)


def deserialize_result(obj: bytes) -> dict:
    """Deserialize an object using orjson into a python dictionary

    Args:
        obj (bytes): The bytes object to deserialize

    Returns:
        dict: The deserialized object as a dictionary
    """
    return orjson.loads(obj)


def deserialize_from_go(obj: Slice_byte) -> dict:
    """Deserialize a Slice_byte object from go into a python dictionary

    Args:
        obj (Slice_byte): The go object to deserialize

    Returns:
        dict: The deserialized object as a dictionary
    """
    return deserialize_result(bytes(obj))



# ***************************************************************** #
# (C) Copyright IBM Corporation 2024.                               #
#                                                                   #
# The source code for this program is not published or otherwise    #
# divested of its trade secrets, irrespective of what has been      #
# deposited with the U.S. Copyright Office.                         #
# ***************************************************************** #
# SPDX-License-Identifier: Apache-2.0
"""This module contains tests for the top-level client operations"""

from datetime import datetime, timedelta

from kubernetes_lite.client import DynamicClient
from kubernetes_lite.envtest import EnvTest
from kubernetes_lite.envtest.pytest import create_random_namespace
from tests.utils import (
    OBJ_API_VERSION,
    OBJ_KIND,
    generate_test_object,
)


def test_rate_limiting(session_kubernetes_env: EnvTest):
    """Test that we can enable ratelimiting"""
    client = DynamicClient(session_kubernetes_env.config(), qps=1, burst=1)
    deployment_res = client.resource(OBJ_API_VERSION, OBJ_KIND)
    obj = generate_test_object(create_random_namespace(client))

    apply_count = 0
    stop_time = datetime.now() + timedelta(seconds=3)
    while True:
        if datetime.now() >= stop_time:
            break
        deployment_res.apply(obj, field_manager="test")
        apply_count += 1

    # Assert that in 3 seconds with a limit of 1 qps we only do a couple of applys
    assert apply_count >= 2
    assert apply_count < 5


def test_no_rate_limiting(session_kubernetes_env: EnvTest):
    """Test that the client has no-ratelimiting by default"""
    client = DynamicClient(session_kubernetes_env.config())
    deployment_res = client.resource(OBJ_API_VERSION, OBJ_KIND)
    obj = generate_test_object(create_random_namespace(client))

    apply_count = 0
    stop_time = datetime.now() + timedelta(seconds=3)
    while True:
        if datetime.now() >= stop_time:
            break
        deployment_res.apply(obj, field_manager="test")
        apply_count += 1

    # Assert that in 3 seconds with a limit of 1 qps we only do a couple of applys
    assert apply_count > 5



# ***************************************************************** #
# (C) Copyright IBM Corporation 2024.                               #
#                                                                   #
# The source code for this program is not published or otherwise    #
# divested of its trade secrets, irrespective of what has been      #
# deposited with the U.S. Copyright Office.                         #
# ***************************************************************** #
# SPDX-License-Identifier: Apache-2.0
"""This module contains utility functions used throughout kubernetes_lite
testing
"""

import copy
from contextlib import nullcontext

from kubernetes_lite.client import DynamicResource
from kubernetes_lite.errors import NotFoundError

import pytest
from deepdiff import DeepDiff

DEFAULT_OBJ_NAME = "test-deployment"
OBJ_KIND = "Deployment"
OBJ_API_VERSION = "apps/v1"


def generate_test_object(namespace: str, name: str = DEFAULT_OBJ_NAME) -> dict:
    """Create a generic kube Deployment to be used for testing

    Args:
        namespace (str): The namespace the resource should be in
        name (str, optional): The name of the object. Defaults to DEFAULT_OBJ_NAME.

    Returns:
        dict: A example deployment object
    """
    return {
        "apiVersion": OBJ_API_VERSION,
        "kind": OBJ_KIND,
        "metadata": {"name": name, "namespace": namespace, "labels": {"app": "test"}},
        "spec": {
            "replicas": 3,
            "selector": {"matchLabels": {"app": "test"}},
            "template": {
                "metadata": {"labels": {"app": "test"}},
                "spec": {"containers": [{"name": "test", "image": "test:1", "ports": [{"containerPort": 80}]}]},
            },
        },
    }


def clean_object_metadata(obj: dict) -> dict:
    """Duplicate an object and remove metadata that changes on every
    request

    Args:
        obj (dict): The object to clean

    Returns:
        dict: A cleaned object
    """
    copied_obj = copy.deepcopy(obj)
    copied_obj["metadata"].pop("resourceVersion", None)
    copied_obj["metadata"].pop("generation", None)
    copied_obj["metadata"].pop("managedFields", None)
    copied_obj["metadata"].pop("uid", None)
    return copied_obj


def validate_result(res_obj: dict | None, expected_obj: dict | None):
    """Validate that two objects are equal

    Args:
        res_obj (dict | None): The result object
        expected_obj (dict | None): The expected object it should match
    """
    assert res_obj
    assert isinstance(res_obj, dict)
    assert expected_obj
    assert isinstance(expected_obj, dict)

    res_obj = clean_object_metadata(res_obj)
    expected_obj = clean_object_metadata(expected_obj)

    diff = DeepDiff(expected_obj, res_obj).to_dict()
    assert len(diff.get("values_changed", [])) == 0
    assert len(diff.get("dictionary_item_removed", [])) == 0


def check_object_status(
    res: DynamicResource,
    namespace: str | None = None,
    name: str = DEFAULT_OBJ_NAME,
    exists: bool = True,
    expected_obj: dict | None = None,
):
    """Check that the object generated by generate_test_object has a certain state in the cluster

    Args:
        res (DynamicResource): The dynamic resource to make requests with
        namespace (str | None, optional): Namespace the object resides in. Defaults to None.
        name (str, optional): Name of the object. Defaults to DEFAULT_OBJ_NAME.
        exists (bool, optional): If the object should exist in the cluster. Useful for delete tests. Defaults to True.
        expected_obj (dict | None, optional): What the cluster object should look like. Defaults to None.
    """
    # Get name from the object
    if not namespace and expected_obj:
        namespace = expected_obj["metadata"].get("namespace")
        name = expected_obj["metadata"].get("name")

    # If the object should exist or not
    error_catcher = nullcontext()
    if not exists:
        error_catcher = pytest.raises(NotFoundError)

    obj = None
    with error_catcher:
        obj = res.get(name=name, namespace=namespace)

    if exists:
        assert obj
        if expected_obj:
            # Remove resourceVersion since those changes do not matter to diff
            validate_result(obj, expected_obj)
    else:
        assert not obj



# ***************************************************************** #
# (C) Copyright IBM Corporation 2024.                               #
#                                                                   #
# The source code for this program is not published or otherwise    #
# divested of its trade secrets, irrespective of what has been      #
# deposited with the U.S. Copyright Office.                         #
# ***************************************************************** #
# SPDX-License-Identifier: Apache-2.0
"""This module contains the implementation of ClientInterface using official kubernetes
client
"""

from typing import Any

from performance.interfaces.base import ClientInterface

import urllib3

# Disable warning for rest to normalize ssl/non-ssl speeds
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


class KubernetesInterface(ClientInterface):
    """KubernetesInterface is a client interface implementation built using the official kubernetes
    library provided at https://github.com/kubernetes-client/python/tree/master
    """

    @staticmethod
    def import_modules() -> list[str]:
        """Import the kubernetes config and dynamic objects"""
        from kubernetes.config import new_client_from_config  # noqa: F401, PLC0415
        from kubernetes.dynamic import DynamicClient  # noqa: F401, PLC0415

        return ["kubernetes"]

    @staticmethod
    def create_client() -> Any:
        """Create the client from a local config"""
        from kubernetes.config import new_client_from_config  # noqa: PLC0415
        from kubernetes.dynamic import DynamicClient  # noqa: PLC0415

        return DynamicClient(new_client_from_config())

    @staticmethod
    def create_resource_client(client: Any, kind: str, api_version: str) -> Any:
        """Create resource client gets the client for a specific resource using the discoverer"""
        return client.resources.get(kind=kind, api_version=api_version)

    @staticmethod
    def create(resource_client: Any, data: Any) -> Any:
        """Create the resource using kubernete's create api"""
        return resource_client.create(data).to_dict()

    @staticmethod
    def apply(resource_client: Any, data: Any) -> Any:
        """Apply the resource using the kubernetes server_side_apply API"""
        return resource_client.server_side_apply(data, field_manager="python").to_dict()

    @staticmethod
    def list(resource_client: Any, namespace: str) -> Any:
        """List all objects in the namespace using kubernetes's generic get(name=None) API"""
        return resource_client.get(name=None, namespace=namespace).to_dict()

    @staticmethod
    def get(resource_client: Any, name: str, namespace: str) -> Any:
        """Get an object from the cluster using the kubernetes resource client and provided name/namespace"""
        return resource_client.get(name=name, namespace=namespace).to_dict()

    @staticmethod
    def delete(resource_client: Any, name: str, namespace: str):
        """Delete an object from the cluster using the kubernetes resource client"""
        return resource_client.delete(name=name, namespace=namespace)



# ***************************************************************** #
# (C) Copyright IBM Corporation 2024.                               #
#                                                                   #
# The source code for this program is not published or otherwise    #
# divested of its trade secrets, irrespective of what has been      #
# deposited with the U.S. Copyright Office.                         #
# ***************************************************************** #
# SPDX-License-Identifier: Apache-2.0
"""This module contains the implementation of the top level dynamic client"""

import datetime
from functools import cached_property
from io import BytesIO

from kubernetes_lite.client.discovery import Discoverer
from kubernetes_lite.client.resource import DynamicResource
from kubernetes_lite.serialization import serialize_to_go
from kubernetes_lite.wrapper.client import (
    NewWrappedDynamicClient,
    NewWrappedDynamicClientWithConfig,
)


class DynamicClient:
    """DynamicClient is the top level client and generic controls settings like authentication,
    rate limiting, and timeouts. Use the .resource method to create a DynamicResource which
    can be used to interact with the cluster.
    """

    def __init__(
        self,
        config: bytes | BytesIO | None = None,
        qps: float = -1,
        burst: int = 100,
        timeout: datetime.timedelta | str = "",
        **kwargs,
    ):
        """Initialize the dynamic client with optional overrides

        Args:
            config (bytes | BytesIO | None, optional): The raw kube_config data to initialize the client with. Defaults
                to controller-runtime's GetConfig.
                See https://pkg.go.dev/sigs.k8s.io/controller-runtime/pkg/client/config#GetConfig for more information.
            qps (float, optional): QPS to set for the client. See https://pkg.go.dev/k8s.io/client-go/rest#Config
                for more information. Defaults to -1 which means rate-limiting is disabled.
            burst (int, optional): How many requests can the client burst to. See https://pkg.go.dev/k8s.io/client-go/rest#Config
                for more information. Defaults to 100.
            timeout (datetime.timedelta | str, optional): Timeout value for the client. If a string is provided check
                https://pkg.go.dev/time#ParseDuration for parsing information. Defaults to "" or no timeout.
            **kwargs: Unused kwargs for backwards compatibility with kubernetes library
        """
        # Parse timeout into a string
        if isinstance(timeout, datetime.timedelta):
            timeout = f"{timeout.total_seconds()}s"

        # Setup the common client arguments
        client_args = [qps, burst, timeout]
        if config:
            self.client = NewWrappedDynamicClientWithConfig(serialize_to_go(config), *client_args)
        else:
            self.client = NewWrappedDynamicClient(*client_args)

    @cached_property
    def resources(self) -> Discoverer:
        """Return the client's Discover object. It's preferred to use the resource()
        function instead of accessing this directly

        Returns:
            Discoverer: The discover used in created DynamicResources
        """
        return Discoverer(self.client)

    def resource(self, api_version: str, kind: str) -> DynamicResource:
        """Return a DynamicResource for the given api_version and kind. This
        DynamicResource can then be used to interact with the cluster.

        Args:
            api_version (str): The api_version for the resource handle
            kind (str): The kind for the resource handle

        Returns:
            DynamicResource: The DynamicResource handle for the requested type
        """
        return self.resources.get(api_version, kind)



# ***************************************************************** #
# (C) Copyright IBM Corporation 2024.                               #
#                                                                   #
# The source code for this program is not published or otherwise    #
# divested of its trade secrets, irrespective of what has been      #
# deposited with the U.S. Copyright Office.                         #
# ***************************************************************** #
# SPDX-License-Identifier: Apache-2.0
"""This module contains imports required by pytest for fixtures"""

from collections.abc import Generator

from kubernetes_lite.client import DynamicClient, DynamicResource
from kubernetes_lite.envtest.pytest import *  # noqa: F403
from tests.utils import OBJ_API_VERSION, OBJ_KIND

import pytest


@pytest.fixture
def resource_namespace(
    session_client_namespace: tuple[DynamicClient, str],
) -> Generator[tuple[DynamicResource, str], None, None]:
    """Helper fixture to return a deployment resource compatible with generate_test_object and the
    namespace to use for tests

    Args:
        session_client_namespace (tuple[DynamicClient, str]): The client and namespace to use

    Yields:
        Generator[tuple[DynamicResource, str], None, None]: Tuple of the dynamic resource and namespace
    """
    client, namespace = session_client_namespace
    yield client.resource(OBJ_API_VERSION, OBJ_KIND), namespace



# ***************************************************************** #
# (C) Copyright IBM Corporation 2024.                               #
#                                                                   #
# The source code for this program is not published or otherwise    #
# divested of its trade secrets, irrespective of what has been      #
# deposited with the U.S. Copyright Office.                         #
# ***************************************************************** #
# SPDX-License-Identifier: Apache-2.0
"""This module contains a list of possible kubernetes errors and helper
functions to convert go_wrapper errors into their respective python
classes
"""

# Assisted by watsonx Code Assistant
from collections.abc import Callable
from typing import ParamSpec, TypeVar


# Copy of all kubernetes errors
class UnknownError(RuntimeError):
    """See https://pkg.go.dev/k8s.io/apimachinery/pkg/apis/meta/v1#StatusReason"""


class UnauthorizedError(RuntimeError):
    """See https://pkg.go.dev/k8s.io/apimachinery/pkg/apis/meta/v1#StatusReason"""


class ForbiddenError(RuntimeError):
    """See https://pkg.go.dev/k8s.io/apimachinery/pkg/apis/meta/v1#StatusReason"""


class NotFoundError(RuntimeError):
    """See https://pkg.go.dev/k8s.io/apimachinery/pkg/apis/meta/v1#StatusReason"""


class AlreadyExistsError(RuntimeError):
    """See https://pkg.go.dev/k8s.io/apimachinery/pkg/apis/meta/v1#StatusReason"""


class ConflictError(RuntimeError):
    """See https://pkg.go.dev/k8s.io/apimachinery/pkg/apis/meta/v1#StatusReason"""


class GoneError(RuntimeError):
    """See https://pkg.go.dev/k8s.io/apimachinery/pkg/apis/meta/v1#StatusReason"""


class InvalidError(RuntimeError):
    """See https://pkg.go.dev/k8s.io/apimachinery/pkg/apis/meta/v1#StatusReason"""


class ServerTimeoutError(RuntimeError):
    """See https://pkg.go.dev/k8s.io/apimachinery/pkg/apis/meta/v1#StatusReason"""


class StoreReadErrorError(RuntimeError):
    """See https://pkg.go.dev/k8s.io/apimachinery/pkg/apis/meta/v1#StatusReason"""


class ClientTimeoutError(RuntimeError):
    """See https://pkg.go.dev/k8s.io/apimachinery/pkg/apis/meta/v1#StatusReason.
    This class corresponds to the general TimeoutError
    """


class TooManyRequestsError(RuntimeError):
    """See https://pkg.go.dev/k8s.io/apimachinery/pkg/apis/meta/v1#StatusReason"""


class BadRequestError(RuntimeError):
    """See https://pkg.go.dev/k8s.io/apimachinery/pkg/apis/meta/v1#StatusReason"""


class MethodNotAllowedError(RuntimeError):
    """See https://pkg.go.dev/k8s.io/apimachinery/pkg/apis/meta/v1#StatusReason"""


class NotAcceptableError(RuntimeError):
    """See https://pkg.go.dev/k8s.io/apimachinery/pkg/apis/meta/v1#StatusReason"""


class RequestEntityTooLargeError(RuntimeError):
    """See https://pkg.go.dev/k8s.io/apimachinery/pkg/apis/meta/v1#StatusReason"""


class UnsupportedMediaTypeError(RuntimeError):
    """See https://pkg.go.dev/k8s.io/apimachinery/pkg/apis/meta/v1#StatusReason"""


class InternalErrorError(RuntimeError):
    """See https://pkg.go.dev/k8s.io/apimachinery/pkg/apis/meta/v1#StatusReason"""


class ExpiredError(RuntimeError):
    """See https://pkg.go.dev/k8s.io/apimachinery/pkg/apis/meta/v1#StatusReason"""


class ServiceUnavailableError(RuntimeError):
    """See https://pkg.go.dev/k8s.io/apimachinery/pkg/apis/meta/v1#StatusReason"""


# Map the kubernetes StatusReason strings to the python errors
EXCEPTION_MAPPING = {
    "Unknown": UnknownError,
    "Unauthorized": UnauthorizedError,
    "Forbidden": ForbiddenError,
    "NotFound": NotFoundError,
    "AlreadyExists": AlreadyExistsError,
    "Conflict": ConflictError,
    "Gone": GoneError,
    "Invalid": InvalidError,
    "ServerTimeout": ServerTimeoutError,
    "StoreReadError": StoreReadErrorError,
    "Timeout": ClientTimeoutError,
    "TooManyRequests": TooManyRequestsError,
    "BadRequest": BadRequestError,
    "MethodNotAllowed": MethodNotAllowedError,
    "NotAcceptable": NotAcceptableError,
    "RequestEntityTooLarge": RequestEntityTooLargeError,
    "UnsupportedMediaType": UnsupportedMediaTypeError,
    "InternalError": InternalErrorError,
    "Expired": ExpiredError,
    "ServiceUnavailable": ServiceUnavailableError,
}


# Vars used in typing the client
KUBE_RETURN_TYPE = TypeVar("KUBE_RETURN_TYPE")
KUBE_PARAM_TYPE = ParamSpec("KUBE_PARAM_TYPE")


def wrap_kube_error(func: Callable[KUBE_PARAM_TYPE, KUBE_RETURN_TYPE]) -> Callable[KUBE_PARAM_TYPE, KUBE_RETURN_TYPE]:
    """Wrap a a function so that generic gopy RuntimeErrors
    are turned into the appropriate kubernetes error type.

    Args:
        func (Callable): The function to wrap

    Returns:
        Callable: The wrapped function
    """

    def wrapped_callable(*args: KUBE_PARAM_TYPE.args, **kwargs: KUBE_PARAM_TYPE.kwargs) -> KUBE_RETURN_TYPE:
        try:
            return func(*args, **kwargs)
        except RuntimeError as exc:
            # If the error has an args attribute then try to parse it into one of
            # the above classes
            if len(exc.args) > 0:
                # Split the string at : to try to extract the message type. This is
                # defined by KubernetesErrorSep in the go code
                message = str(exc.args[0])
                split_message = message.split(":")

                # Use the first part of the message to map to the exception type
                if split_message[0] in EXCEPTION_MAPPING:
                    new_exc_type = EXCEPTION_MAPPING[split_message[0]]

                    # Recreate the unsplit message with out the parsed bit
                    exc_message = ":".join(split_message[1:])
                    raise new_exc_type(exc_message) from None
            raise exc

    return wrapped_callable



# ***************************************************************** #
# (C) Copyright IBM Corporation 2024.                               #
#                                                                   #
# The source code for this program is not published or otherwise    #
# divested of its trade secrets, irrespective of what has been      #
# deposited with the U.S. Copyright Office.                         #
# ***************************************************************** #
# SPDX-License-Identifier: Apache-2.0
"""This module contains the DynamicResource which is the main way to interact
with the cluster
"""

# Assisted by watsonx Code Assistant
import warnings
from typing import Any

from kubernetes_lite.client.sanitization import SourceObjectType, process_input_resource
from kubernetes_lite.client.watch import Watch
from kubernetes_lite.errors import wrap_kube_error
from kubernetes_lite.serialization import (
    deserialize_from_go,
    serialize_to_go,
)
from kubernetes_lite.wrapper.client import (
    WrappedNamespaceableResourceInterface,
)


class DynamicResource:
    """DynamicResource is the primary client for interacting with resources on the cluster. It's
    functions mirror the ResourceInterface provided by https://pkg.go.dev/k8s.io/client-go/dynamic#ResourceInterface.
    The parameters to the various functions map to the json names of the Options objects in
    https://pkg.go.dev/k8s.io/apimachinery/pkg/apis/meta/v1. For example `list()` parameter's map
    to ListOptions
    """

    def __init__(self, resource: WrappedNamespaceableResourceInterface):
        """Create a DynamicResource which points to a WrappedNamespaceableResourceInterface from go. This class
        should not be initialized directly and should instead be created by the client/discoverer

        Args:
            resource (WrappedNamespaceableResourceInterface): The WrappedNamespaceableResourceInterface to wrap
        """
        self._resource = resource

    @wrap_kube_error
    def get(
        self,
        name: str | None = None,
        namespace: str | None = None,
        resource_version: str | None = None,
        extra_options_kwargs: dict[str, Any] | None = None,
        subresources: list[str] | None = None,
        **kwargs,
    ) -> dict:
        """Get an object from the cluster. The parameters map to https://pkg.go.dev/k8s.io/apimachinery/pkg/apis/meta/v1#GetOptions

        Args:
            name (str | None, optional): The name of the resource to fetch. **None is Deprecated**. Defaults to None.
            namespace (str | None, optional): Optional namespace for the resource. Defaults to None.
            resource_version (str | None, optional): Optional resourceVersion to fetch. Defaults to None.
            extra_options_kwargs (dict[str, Any] | None, optional): Optional dictionary containing more GetOption
                parameters. Defaults to None.
            subresources (list[str] | None, optional): Subresource of an object to get. Defaults to None.
            **kwargs: Unused kwargs for compatibility

        Returns:
            dict: The object from the cluster as a dictionary
        """
        if name is None:
            warnings.warn(
                "Get with no name to list objects is deprecated. Use the list method instead",
                DeprecationWarning,
                stacklevel=2,
            )
            return self.list(namespace=namespace, resource_version=resource_version, **kwargs)

        options = {**(extra_options_kwargs or {}), **kwargs}
        if resource_version:
            options["resourceVersion"] = resource_version
        serialized_options = serialize_to_go(options)

        resource_ptr = self._resource
        if namespace:
            resource_ptr = resource_ptr.Namespace(namespace)

        subresource_options = subresources if subresources else []
        result = resource_ptr.Get(name, serialized_options, *subresource_options)
        return deserialize_from_go(result)

    @wrap_kube_error
    def create(  # noqa: PLR0917, PLR0913
        self,
        resource: SourceObjectType,
        namespace: str | None = None,
        dry_run: list[str] | None = None,
        field_manager: str | None = None,
        field_validation: str | None = None,
        extra_options_kwargs: dict[str, Any] | None = None,
        subresources: list[str] | None = None,
        **kwargs,
    ) -> dict:
        """Create an object in the cluster. The parameters map to https://pkg.go.dev/k8s.io/apimachinery/pkg/apis/meta/v1#CreateOptions

        Args:
            resource (SourceObjectType): The resource to deploy
            namespace (str | None, optional): Optional namespace for the resource. Defaults to None or namespace
                in the provided resources.
            dry_run (list[str] | None, optional): Optional List of string dryrruns to use. Valid values can be gathered
                from the golang docs. Defaults to None.
            field_manager (str | None, optional): Optional field manager string to use to manage resource fields.
                Defaults to None.
            field_validation (str | None, optional): Optional fieldValidation instructs the server on how to
                handle objects in the request containing unknown or duplicate fields. Defaults to None.
            extra_options_kwargs (dict[str, Any] | None, optional): Optional dictionary containing more CreateOption
                parameters. Defaults to None.
            subresources (list[str] | None, optional): Subresource of an object to get. Defaults to None.
            **kwargs: Unused kwargs for compatibility

        Returns:
            dict: The created object from the cluster in a dictionary
        """
        options = {**(extra_options_kwargs or {}), **kwargs}
        if dry_run:
            options["dryRun"] = dry_run
        if field_manager:
            options["fieldManager"] = field_manager
        if field_validation:
            options["fieldValidation"] = field_validation
        serialized_options = serialize_to_go(options)

        serialized_resource, resource_metadata = process_input_resource(resource)
        if not namespace and resource_metadata.namespace:
            namespace = resource_metadata.namespace
        resource_ptr = self._resource
        if namespace:
            resource_ptr = resource_ptr.Namespace(namespace)

        subresource_options = subresources if subresources else []
        result = resource_ptr.Create(serialized_resource, serialized_options, *subresource_options)
        return deserialize_from_go(result)

    @wrap_kube_error
    def update(  # noqa: PLR0917, PLR0913
        self,
        resource: SourceObjectType,
        namespace: str | None = None,
        dry_run: list[str] | None = None,
        field_manager: str | None = None,
        field_validation: str | None = None,
        extra_options_kwargs: dict[str, Any] | None = None,
        subresources: list[str] | None = None,
        **kwargs,
    ) -> dict:
        """Update an object in the cluster. The parameters map to https://pkg.go.dev/k8s.io/apimachinery/pkg/apis/meta/v1#UpdateOptions

        Args:
            resource (SourceObjectType): The resource to deploy
            namespace (str | None, optional): Optional namespace for the resource. Defaults to None or namespace
                in the provided resources.
            dry_run (list[str] | None, optional): Optional List of string dryrruns to use. Valid values can be gathered
                from the golang docs. Defaults to None.
            field_manager (str | None, optional): Optional field manager string to use to manage resource fields.
                Defaults to None.
            field_validation (str | None, optional): Optional fieldValidation instructs the server on how to
                handle objects in the request containing unknown or duplicate fields. Defaults to None.
            extra_options_kwargs (dict[str, Any] | None, optional): Optional dictionary containing more CreateOption
                parameters. Defaults to None.
            subresources (list[str] | None, optional): Subresource of an object to get. Defaults to None.
            **kwargs: Unused kwargs for compatibility

        Returns:
            dict: The updated object from the cluster in a dictionary
        """
        options = {**(extra_options_kwargs or {}), **kwargs}
        if dry_run:
            options["dryRun"] = dry_run
        if field_manager:
            options["fieldManager"] = field_manager
        if field_validation:
            options["fieldValidation"] = field_validation
        serialized_options = serialize_to_go(options)

        serialized_resource, resource_metadata = process_input_resource(resource)
        if not namespace and resource_metadata.namespace:
            namespace = resource_metadata.namespace
        resource_ptr = self._resource
        if namespace:
            resource_ptr = resource_ptr.Namespace(namespace)

        subresource_options = subresources if subresources else []
        result = resource_ptr.Update(serialized_resource, serialized_options, *subresource_options)
        return deserialize_from_go(result)

    @wrap_kube_error
    def update_status(  # noqa: PLR0917, PLR0913
        self,
        resource: SourceObjectType,
        namespace: str | None = None,
        dry_run: list[str] | None = None,
        field_manager: str | None = None,
        field_validation: str | None = None,
        extra_options_kwargs: dict[str, Any] | None = None,
        subresources: list[str] | None = None,
        **kwargs,
    ) -> dict:
        """Update the status object in the cluster. The parameters map to https://pkg.go.dev/k8s.io/apimachinery/pkg/apis/meta/v1#UpdateOptions

        Args:
            resource (SourceObjectType): The resource to deploy
            namespace (str | None, optional): Optional namespace for the resource. Defaults to None or namespace
                in the provided resources.
            dry_run (list[str] | None, optional): Optional List of string dryrruns to use. Valid values can be gathered
                from the golang docs. Defaults to None.
            field_manager (str | None, optional): Optional field manager string to use to manage resource fields.
                Defaults to None.
            field_validation (str | None, optional): Optional fieldValidation instructs the server on how to
                handle objects in the request containing unknown or duplicate fields. Defaults to None.
            extra_options_kwargs (dict[str, Any] | None, optional): Optional dictionary containing more CreateOption
                parameters. Defaults to None.
            subresources (list[str] | None, optional): Subresource of an object to get. Defaults to None.
            **kwargs: Unused kwargs for compatibility

        Returns:
            dict: The updated object from the cluster in a dictionary
        """
        options = {**(extra_options_kwargs or {}), **kwargs}
        if dry_run:
            options["dryRun"] = dry_run
        if field_manager:
            options["fieldManager"] = field_manager
        if field_validation:
            options["fieldValidation"] = field_validation
        serialized_options = serialize_to_go(options)

        serialized_resource, resource_metadata = process_input_resource(resource)
        if not namespace and resource_metadata.namespace:
            namespace = resource_metadata.namespace
        resource_ptr = self._resource
        if namespace:
            resource_ptr = resource_ptr.Namespace(namespace)

        subresource_options = subresources if subresources else []
        result = resource_ptr.UpdateStatus(serialized_resource, serialized_options, *subresource_options)
        return deserialize_from_go(result)

    @wrap_kube_error
    def patch(  # noqa: PLR0917, PLR0913
        self,
        name: str,
        patch_type: str,
        patch_data: SourceObjectType,
        namespace: str | None = None,
        dry_run: list[str] | None = None,
        field_manager: str | None = None,
        field_validation: str | None = None,
        force: bool | None = None,
        extra_options_kwargs: dict[str, Any] | None = None,
        subresources: list[str] | None = None,
        **kwargs,
    ) -> dict:
        """Patch an object in the cluster. The parameters map to https://pkg.go.dev/k8s.io/apimachinery/pkg/apis/meta/v1#PatchOptions

        Args:
            name (str): Name of the object to patch
            patch_type (str): Type of patch to use. See https://pkg.go.dev/k8s.io/apimachinery/pkg/types#PatchType for
                available options
            patch_data (SourceObjectType): The raw patch data to use

            resource (SourceObjectType): The resource to deploy
            namespace (str | None, optional): Optional namespace for the resource. Defaults to None or namespace
                in the provided resources.
            dry_run (list[str] | None, optional): Optional List of string dryrruns to use. Valid values can be gathered
                from the golang docs. Defaults to None.
            field_manager (str | None, optional): Optional field manager string to use to manage resource fields.
                Defaults to None.
            force (bool | None, optional): If the patch should be forced ignoring field_manager. Defaults to None.
            field_validation (str | None, optional): Optional fieldValidation instructs the server on how to
                handle objects in the request containing unknown or duplicate fields. Defaults to None.
            extra_options_kwargs (dict[str, Any] | None, optional): Optional dictionary containing more PatchOptions
                parameters. Defaults to None.
            subresources (list[str] | None, optional): Subresource of an object to get. Defaults to None.
            **kwargs: Unused kwargs for compatibility

        Returns:
            dict: The patched object from the cluster in a dictionary
        """
        options = {**(extra_options_kwargs or {}), **kwargs}
        if dry_run:
            options["dryRun"] = dry_run
        if field_manager:
            options["fieldManager"] = field_manager
        if field_validation:
            options["fieldValidation"] = field_validation
        if field_manager:
            options["force"] = force
        serialized_options = serialize_to_go(options)

        serialized_patch = serialize_to_go(patch_data)
        resource_ptr = self._resource
        if namespace:
            resource_ptr = resource_ptr.Namespace(namespace)

        subresource_options = subresources if subresources else []
        result = resource_ptr.Patch(name, patch_type, serialized_patch, serialized_options, *subresource_options)
        return deserialize_from_go(result)

    @wrap_kube_error
    def apply(  # noqa: PLR0917, PLR0913
        self,
        resource: SourceObjectType,
        name: str | None = None,
        namespace: str | None = None,
        dry_run: bool | None = None,
        field_manager: str | None = None,
        force: bool | None = None,
        extra_options_kwargs: dict[str, Any] | None = None,
        subresources: list[str] | None = None,
        **kwargs,
    ) -> dict:
        """Apply an object in the cluster. The parameters map to https://pkg.go.dev/k8s.io/apimachinery/pkg/apis/meta/v1#ApplyOptions.
        See https://kubernetes.io/docs/reference/using-api/server-side-apply/ for more information

        Args:
            resource (SourceObjectType): The resource to deploy
            name (str | None, optional): Optional name for the resource. Defaults to None or name
                in the provided resources.
            namespace (str | None, optional): Optional namespace for the resource. Defaults to None or namespace
                in the provided resources.
            dry_run (list[str] | None, optional): Optional List of string dryrruns to use. Valid values can be gathered
                from the golang docs. Defaults to None.
            field_manager (str | None, optional): Optional field manager string to use to manage resource fields.
                Defaults to None.
            force (bool | None, optional): If the patch should be forced ignoring field_manager. Defaults to None.
            extra_options_kwargs (dict[str, Any] | None, optional): Optional dictionary containing more ApplyOptions
                parameters. Defaults to None.
            subresources (list[str] | None, optional): Subresource of an object to get. Defaults to None.
            **kwargs: Unused kwargs for compatibility

        Returns:
            dict: The applied object from the cluster in a dictionary
        """
        options = {**(extra_options_kwargs or {}), **kwargs}
        if dry_run:
            options["dryRun"] = dry_run
        if field_manager:
            options["fieldManager"] = field_manager
        if force:
            options["force"] = force
        serialized_options = serialize_to_go(options)

        serialized_resource, resource_metadata = process_input_resource(resource)
        if not namespace and resource_metadata.namespace:
            namespace = resource_metadata.namespace
        if not name and resource_metadata.name:
            name = resource_metadata.name

        resource_ptr = self._resource
        if namespace:
            resource_ptr = resource_ptr.Namespace(namespace)

        subresource_options = subresources if subresources else []
        result = resource_ptr.Apply(name, serialized_resource, serialized_options, *subresource_options)
        return deserialize_from_go(result)

    @wrap_kube_error
    def apply_status(  # noqa: PLR0917, PLR0913
        self,
        resource: SourceObjectType,
        name: str | None = None,
        namespace: str | None = None,
        dry_run: bool | None = None,
        field_manager: str | None = None,
        force: bool | None = None,
        extra_options_kwargs: dict[str, Any] | None = None,
        subresources: list[str] | None = None,
        **kwargs,
    ) -> dict:
        """Apply a status object in the cluster. The parameters map to https://pkg.go.dev/k8s.io/apimachinery/pkg/apis/meta/v1#ApplyOptions.
        See https://kubernetes.io/docs/reference/using-api/server-side-apply/ for more information

        Args:
            resource (SourceObjectType): The resource to deploy
            name (str | None, optional): Optional name for the resource. Defaults to None or name
                in the provided resources.
            namespace (str | None, optional): Optional namespace for the resource. Defaults to None or namespace
                in the provided resources.
            dry_run (list[str] | None, optional): Optional List of string dryrruns to use. Valid values can be gathered
                from the golang docs. Defaults to None.
            field_manager (str | None, optional): Optional field manager string to use to manage resource fields.
                Defaults to None.
            force (bool | None, optional): If the patch should be forced ignoring field_manager. Defaults to None.
            field_validation (str | None, optional): Optional fieldValidation instructs the server on how to
                handle objects in the request containing unknown or duplicate fields. Defaults to None.
            extra_options_kwargs (dict[str, Any] | None, optional): Optional dictionary containing more ApplyOptions
                parameters. Defaults to None.
            subresources (list[str] | None, optional): Subresource of an object to get. Defaults to None.
            **kwargs: Unused kwargs for compatibility

        Returns:
            dict: The applied object from the cluster in a dictionary
        """
        options = {**(extra_options_kwargs or {}), **kwargs}
        if dry_run:
            options["dryRun"] = dry_run
        if field_manager:
            options["fieldManager"] = field_manager
        if force:
            options["force"] = force
        serialized_options = serialize_to_go(options)

        serialized_resource, resource_metadata = process_input_resource(resource)
        if not namespace and resource_metadata.namespace:
            namespace = resource_metadata.namespace
        if not name and resource_metadata.name:
            name = resource_metadata.name

        resource_ptr = self._resource
        if namespace:
            resource_ptr = resource_ptr.Namespace(namespace)

        subresource_options = subresources if subresources else []
        result = resource_ptr.ApplyStatus(name, serialized_resource, serialized_options, *subresource_options)
        return deserialize_from_go(result)

    @wrap_kube_error
    def delete(  # noqa: PLR0917, PLR0913
        self,
        name: str,
        namespace: str | None = None,
        grace_period_seconds: int | None = None,
        propagation_policy: str | None = None,
        dry_run: list[str] | None = None,
        ignore_store_read_error: bool | None = None,
        extra_options_kwargs: dict[str, Any] | None = None,
        **kwargs,
    ):
        """Delete an object from the cluster. The parameters map to https://pkg.go.dev/k8s.io/apimachinery/pkg/apis/meta/v1#DeleteOptions.

        Args:
            name (str): Name for the resource to delete.
            namespace (str | None, optional): Optional namespace for the resource. Defaults to None
            dry_run (list[str] | None, optional): Optional List of string dryrruns to use. Valid values can be gathered
                from the golang docs. Defaults to None.
            grace_period_seconds (int | None, optional): Optional time to wait for graceful termination before
                deleting the resource. Defaults to None.
            propagation_policy (str | None, optional): Optional DeletionPropagation decides how a deletion propagates
                to dependents of the object. See https://pkg.go.dev/k8s.io/apimachinery/pkg/apis/meta/v1#DeletionPropagation
                for more information. Defaults to None.
            ignore_store_read_error (bool | None, optional): Optional override to force successful deletion even
                if there is an etcd store error. Defaults to None.
            extra_options_kwargs (dict[str, Any] | None, optional): Optional dictionary containing more DeleteOptions
                parameters. Defaults to None.
            subresources (list[str] | None, optional): Subresource of an object to get. Defaults to None.
            **kwargs: Unused kwargs for compatibility
        """
        options = {**(extra_options_kwargs or {}), **kwargs}
        if grace_period_seconds:
            options["gracePeriodSeconds"] = grace_period_seconds
        if propagation_policy:
            options["PropagationPolicy"] = propagation_policy
        if dry_run:
            options["dryRun"] = dry_run
        if ignore_store_read_error:
            options["IgnoreStoreReadErrorWithClusterBreakingPotential"] = ignore_store_read_error
        serialized_options = serialize_to_go(options)

        resource_ptr = self._resource
        if namespace:
            resource_ptr = resource_ptr.Namespace(namespace)

        resource_ptr.Delete(name, serialized_options)

    @wrap_kube_error
    def delete_collection(  # noqa: PLR0917, PLR0913
        self,
        namespace: str | None = None,
        grace_period_seconds: int | None = None,
        propagation_policy: str | None = None,
        dry_run: list[str] | None = None,
        limit: int | None = None,
        resource_version: str | None = None,
        resource_version_match: str | None = None,
        label_selector: str | None = None,
        field_selector: str | None = None,
        timeout: int | None = None,
        ignore_store_read_error: bool | None = None,
        extra_delete_options_kwargs: dict[str, Any] | None = None,
        extra_list_options_kwargs: dict[str, Any] | None = None,
        **kwargs,
    ):
        """Delete a collection of objects from the cluster. The parameters map to both
        https://pkg.go.dev/k8s.io/apimachinery/pkg/apis/meta/v1#DeleteOptions and
         https://pkg.go.dev/k8s.io/apimachinery/pkg/apis/meta/v1#ListOptions

        Args:
            name (str): Name for the resource to delete.
            namespace (str | None, optional): Optional namespace for the resource. Defaults to None
            dry_run (list[str] | None, optional): Optional List of string dryrruns to use. Valid values can be gathered
                from the golang docs. Defaults to None.
            grace_period_seconds (int | None, optional): Optional time to wait for graceful termination before
                deleting the resource. Defaults to None.
            propagation_policy (str | None, optional): Optional DeletionPropagation decides how a deletion propagates
                to dependents of the object. See https://pkg.go.dev/k8s.io/apimachinery/pkg/apis/meta/v1#DeletionPropagation
                for more information. Defaults to None.
            ignore_store_read_error (bool | None, optional): Optional override to force successful deletion even
                if there is an etcd store error. Defaults to None.
            extra_delete_options_kwargs (dict[str, Any] | None, optional): Optional dictionary containing more
                DeleteOptions parameters. Defaults to None.
            resource_version (str | None, optional): Optional resourceVersion to fetch. Maps to the resourceVersion of
                the ObjectList object. Defaults to None.
            resource_version_match (str | None, optional): Optional regex string for filtering returned resources.
                Defaults to None.
            label_selector (str | None, optional): Optional label selector to filter returned resources.
                Defaults to None.
            field_selector (str | None, optional): Optional field selector to filter returned resources.
                Defaults to None.
            timeout (int | None, optional): Optional server timeout for the request. Defaults to None.
            limit (int | None, optional): Optional max number of resources returned. Defaults to None.
            extra_list_options_kwargs (dict[str, Any] | None, optional): Optional dictionary containing more ListOption
                parameters. Defaults to None.
            **kwargs: Unused kwargs for compatibility
        """
        delete_options = {**(extra_delete_options_kwargs or {}), **kwargs}
        if grace_period_seconds:
            delete_options["gracePeriodSeconds"] = grace_period_seconds
        if propagation_policy:
            delete_options["PropagationPolicy"] = propagation_policy
        if dry_run:
            delete_options["dryRun"] = dry_run
        if ignore_store_read_error:
            delete_options["IgnoreStoreReadErrorWithClusterBreakingPotential"] = ignore_store_read_error

        list_options = {**(extra_list_options_kwargs or {}), **kwargs}
        if resource_version:
            list_options["resourceVersion"] = resource_version
        if resource_version_match:
            list_options["resourceVersionMatch"] = resource_version_match
        if label_selector:
            list_options["labelSelector"] = label_selector
        if field_selector:
            list_options["fieldSelector"] = field_selector
        if timeout:
            list_options["timeoutSeconds"] = timeout
        if limit:
            list_options["Limit"] = limit
        serialized_del_options = serialize_to_go(delete_options)
        serialized_list_options = serialize_to_go(list_options)

        resource_ptr = self._resource
        if namespace:
            resource_ptr = resource_ptr.Namespace(namespace)
        resource_ptr.DeleteCollection(serialized_del_options, serialized_list_options)

    @wrap_kube_error
    def watch(  # noqa: PLR0917, PLR0913
        self,
        namespace: str | None = None,
        resource_version: str | None = None,
        resource_version_match: str | None = None,
        label_selector: str | None = None,
        field_selector: str | None = None,
        timeout: int | None = None,
        limit: int | None = None,
        continue_req: str | None = None,
        extra_options_kwargs: dict[str, Any] | None = None,
        watcher: Watch | None = None,
        **kwargs,
    ) -> Watch:
        """Watch for events from the cluster. The parameters map to https://pkg.go.dev/k8s.io/apimachinery/pkg/apis/meta/v1#WatchOptions

        Args:
            namespace (str | None, optional): Optional namespace for the resource. Defaults to None.
            resource_version (str | None, optional): Optional resourceVersion to fetch. Maps to the resourceVersion of
                the ObjectList object. Defaults to None.
            resource_version_match (str | None, optional): Optional regex string for filtering returned resources.
                Defaults to None.
            label_selector (str | None, optional): Optional label selector to filter returned resources.
                Defaults to None.
            field_selector (str | None, optional): Optional field selector to filter returned resources.
                Defaults to None.
            timeout (int | None, optional): Optional server timeout for the request. Defaults to None.
            limit (int | None, optional): Optional max number of resources returned. Defaults to None.
            continue_req (str | None, optional): Optional continue token to use to iterate over more results. Used with
                limit keywarg. Defaults to None.
            extra_options_kwargs (dict[str, Any] | None, optional): Optional dictionary containing more ListOption
                parameters. Defaults to None.
            subresources (list[str] | None, optional): Optional subresource of an object to get. Defaults to None.
            watcher (Watch | None, optional): Optional Watch object to use as the watcher interface. Users
                can use this to stop the event stream during or outside of iteration
            **kwargs: Unused kwargs for compatibility

        Returns:
            dict: The ObjectList object from the cluster as a dictionary e.g. use `.get("items")` to get list of
                resources
        """
        options = {**(extra_options_kwargs or {}), **kwargs}
        if resource_version:
            options["resourceVersion"] = resource_version
        if resource_version_match:
            options["resourceVersionMatch"] = resource_version_match
        if label_selector:
            options["labelSelector"] = label_selector
        if field_selector:
            options["fieldSelector"] = field_selector
        if timeout:
            options["timeoutSeconds"] = timeout
        if limit:
            options["Limit"] = limit
        if continue_req:
            options["continue"] = continue_req
        serialized_options = serialize_to_go(options)

        resource_ptr = self._resource
        if namespace:
            resource_ptr = resource_ptr.Namespace(namespace)

        if not watcher:
            watcher = Watch()

        interface = resource_ptr.Watch(serialized_options)
        watcher.interface = interface

        return watcher

    # ! Defined at bottom of class to avoid overriding built-in list
    @wrap_kube_error
    def list(  # noqa: PLR0917, PLR0913
        self,
        namespace: str | None = None,
        limit: int | None = None,
        resource_version: str | None = None,
        resource_version_match: str | None = None,
        label_selector: str | None = None,
        field_selector: str | None = None,
        timeout: int | None = None,
        continue_req: str | None = None,
        extra_options_kwargs: dict[str, Any] | None = None,
        **kwargs,
    ) -> dict:
        """List an object from the cluster. The parameters map to https://pkg.go.dev/k8s.io/apimachinery/pkg/apis/meta/v1#ListOptions

        Args:
            namespace (str | None, optional): Optional namespace for the resource. Defaults to None.
            resource_version (str | None, optional): Optional resourceVersion to fetch. Maps to the resourceVersion of
                the ObjectList object. Defaults to None.
            resource_version_match (str | None, optional): Optional regex string for filtering returned resources.
                Defaults to None.
            label_selector (str | None, optional): Optional label selector to filter returned resources.
                Defaults to None.
            field_selector (str | None, optional): Optional field selector to filter returned resources.
                Defaults to None.
            timeout (int | None, optional): Optional server timeout for the request. Defaults to None.
            limit (int | None, optional): Optional max number of resources returned. Defaults to None.
            continue_req (str | None, optional): Optional continue token to use to iterate over more results. Used with
                limit keywarg. Defaults to None.
            extra_options_kwargs (dict[str, Any] | None, optional): Optional dictionary containing more ListOption
                parameters. Defaults to None.
            subresources (list[str] | None, optional): Optional subresource of an object to get. Defaults to None.
            **kwargs: Unused kwargs for compatibility

        Returns:
            dict: The ObjectList object from the cluster as a dictionary e.g. use `.get("items")` to get list of
                resources
        """
        options = {
            **(extra_options_kwargs or {}),
            **kwargs,
        }
        if resource_version:
            options["resourceVersion"] = resource_version
        if resource_version_match:
            options["resourceVersionMatch"] = resource_version_match
        if label_selector:
            options["labelSelector"] = label_selector
        if field_selector:
            options["fieldSelector"] = field_selector
        if timeout:
            options["timeoutSeconds"] = timeout
        if limit:
            options["Limit"] = limit
        if continue_req:
            options["continue"] = continue_req
        serialized_options = serialize_to_go(options)

        resource_ptr = self._resource
        if namespace:
            resource_ptr = resource_ptr.Namespace(namespace)
        result = resource_ptr.List(serialized_options)
        return deserialize_from_go(result)



# ***************************************************************** #
# (C) Copyright IBM Corporation 2024.                               #
#                                                                   #
# The source code for this program is not published or otherwise    #
# divested of its trade secrets, irrespective of what has been      #
# deposited with the U.S. Copyright Office.                         #
# ***************************************************************** #
# SPDX-License-Identifier: Apache-2.0
"""Kubernetes_lite is a lightweight wrapper around the dynamic client from k8s.io/client-go/dynamic,
the 'mock' kubernetes server from sigs.k8s.io/controller-runtime/pkg/envtest, and its helper tool
setup-envtest https://pkg.go.dev/sigs.k8s.io/controller-runtime/tools/setup-envtest
"""

try:
    # Local
    from ._version import __version__, __version_tuple__  # noqa: F401 # unused import
except ImportError:
    __version__ = "unknown"
    version_tuple = (0, 0, __version__)



# ***************************************************************** #
# (C) Copyright IBM Corporation 2024.                               #
#                                                                   #
# The source code for this program is not published or otherwise    #
# divested of its trade secrets, irrespective of what has been      #
# deposited with the U.S. Copyright Office.                         #
# ***************************************************************** #
# SPDX-License-Identifier: Apache-2.0
"""This module contains the main entrypoint for running performance tests"""

from performance.command import run_test

if __name__ == "__main__":
    import typer

    app = typer.Typer()
    app.command()(run_test)
    app()



# ***************************************************************** #
# (C) Copyright IBM Corporation 2024.                               #
#                                                                   #
# The source code for this program is not published or otherwise    #
# divested of its trade secrets, irrespective of what has been      #
# deposited with the U.S. Copyright Office.                         #
# ***************************************************************** #
# SPDX-License-Identifier: Apache-2.0
"""This is a helper script to install the build system dependencies. For now this
just installs the correct version of golang. Due to where this is ran we can not have
any external dependencies (e.g. requests)
"""

import copy
import os
import platform
import shutil
import subprocess
import sys
import tempfile
from enum import StrEnum
from pathlib import Path
from urllib.request import urlopen

from scripts.utils import ARCH, SYSTEM, SystemTypes

# Base settings for go/src/dest
GO_VERSION = "1.23.7"
GO_BASE_URL = "https://go.dev/dl/go{version!s}.{system!s}-{arch!s}.{archive!s}"
DEST_PATH = Path(tempfile.gettempdir())


class ArchiveType(StrEnum):
    """ArchiveType is a simple enum to track what type of archive go stores for a platform"""

    ZIP = "zip"
    TAR = "tar.gz"


archive = None
if platform.system() == "Linux" or platform.system() == "Darwin":
    archive = ArchiveType.TAR
elif platform.system() == "Windows":
    archive = ArchiveType.ZIP
else:
    raise ValueError(f"Unknown machine system {platform.system()}")

# Setup a directory to download the tar to
with tempfile.TemporaryDirectory() as temp_dir:
    temp_path = Path(temp_dir)
    archive_file = temp_path / f"go.{archive!s}"

    # Template the url and send the request
    go_url = GO_BASE_URL.format(
        version=GO_VERSION,
        arch=ARCH,
        system=SYSTEM,
        archive=archive,
    )
    print(f"Attempting to download go src from: {go_url}", file=sys.stderr)
    with urlopen(go_url) as response:
        archive_file.write_bytes(response.read())

    print(f"Attempting to extract archive to: {DEST_PATH}", file=sys.stderr)
    shutil.unpack_archive(archive_file, str(DEST_PATH))

# Print the path used for installation
go_path = DEST_PATH / "go/bin"
go_bin = go_path / "go"
print(f"Binary path for go: {go_path}", file=sys.stderr)
current_env = copy.deepcopy(os.environ)
current_env["GOBIN"] = str(go_path)

# Ensure go is usable. ! Note this requires the path be set externally
subprocess.run([f"{go_bin}", "version"], check=True, stdout=sys.stderr, env=current_env)
# ! Install the make gen dependencies
subprocess.run(
    ["pip", "install", "pybindgen", "typer"],
    check=True,
    stdout=sys.stderr,
    env=current_env,
)
subprocess.run(
    [f"{go_bin}", "install", "golang.org/x/tools/cmd/goimports@latest"],
    check=True,
    stdout=sys.stderr,
    env=current_env,
)
subprocess.run(
    [f"{go_bin}", "install", "github.com/go-python/gopy@latest"], check=True, stdout=sys.stderr, env=current_env
)


if SYSTEM in {SystemTypes.LINUX, SystemTypes.DARWIN}:
    print(f"export PATH=$PATH:{go_path}")
    print(f"export GO_INSTALL_PATH={go_path}")
else:
    print(f"SET PATH=%PATH%;{go_path}")
    print(f"SET GO_INSTALL_PATH={go_path}")
    print(f"ls {go_path}")

# System-link go binaries to a usr defined path
if SYSTEM in {SystemTypes.DARWIN, SystemTypes.LINUX}:
    print(f"ln -s {go_path}/* /usr/local/bin")



# ***************************************************************** #
# (C) Copyright IBM Corporation 2024.                               #
#                                                                   #
# The source code for this program is not published or otherwise    #
# divested of its trade secrets, irrespective of what has been      #
# deposited with the U.S. Copyright Office.                         #
# ***************************************************************** #
# SPDX-License-Identifier: Apache-2.0
"""This module contains the objects used in discovering GVR's in the cluster"""

import warnings

from kubernetes_lite.client.resource import DynamicResource
from kubernetes_lite.errors import wrap_kube_error
from kubernetes_lite.wrapper.client import (
    WrappedDynamicClient,
)


class Discoverer:
    """Discoverer is used to discover resource kinds/apiVersions and generate the
    corresponding DynamicResource
    """

    def __init__(self, client: WrappedDynamicClient):
        """Initialize the Discoverer with a provided client

        Args:
            client (WrappedDynamicClient): The client to use for interacting with the cluster
        """
        self._client = client

    @wrap_kube_error
    def get(self, api_version: str, kind: str) -> DynamicResource:
        """Return a DynamicResource for the given api_version and kind. This
        DynamicResource can then be used to interact with the cluster.

        Args:
            api_version (str): The api_version for the resource handle
            kind (str): The kind for the resource handle

        Returns:
            DynamicResource: The DynamicResource handle for the requested type
        """
        kind = kind.lower()
        return DynamicResource(self._client.Resource(api_version, kind))

    def search(self, **kwargs) -> DynamicResource:
        """See get for arguments/returns"""
        warnings.warn(
            "The search method on resources/Discoverer is deprecated. Use the get method instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.get(**kwargs)



# ***************************************************************** #
# (C) Copyright IBM Corporation 2024.                               #
#                                                                   #
# The source code for this program is not published or otherwise    #
# divested of its trade secrets, irrespective of what has been      #
# deposited with the U.S. Copyright Office.                         #
# ***************************************************************** #
# SPDX-License-Identifier: Apache-2.0
"""This module contains the implementation fo the pythonic wrapper around
https://pkg.go.dev/sigs.k8s.io/controller-runtime/pkg/envtest
"""

import atexit
from functools import cached_property
from io import BytesIO
from logging import Logger, getLogger
from threading import Event

from kubernetes_lite.client import DynamicClient
from kubernetes_lite.setup_envtest import internal_run_setup_envtest_command
from kubernetes_lite.wrapper.server import EnvTestEnvironment, NewEnvTestEnvironment, NewEnvTestEnvironmentWithPath


class EnvTest:
    """EnvTest is a wrapper around the golang EnvTest object provided by
    https://pkg.go.dev/sigs.k8s.io/controller-runtime/pkg/envtest. This class
    exposes methods to start and stop the server as well as fetch a raw kubeconfig
    or client object. This class can also handle calling setup-envtest to download
    the required binaries and create the path.

    To configure the underyling golang EnvTestEnvironment checkout the environmental variables
    exposed by EnvTest: https://book.kubebuilder.io/reference/envtest.html#environment-variables
    """

    def __init__(self, log: Logger | None = None):
        """Initialize the EnvTest object and create an event to ensure the server isn't
        shutdown multiple times

        Arguments:
            log (logger | None): Optional logger to provide. Defaults to None
        """
        self._log = log if log else getLogger("envtest")
        self._env: EnvTestEnvironment | None = None
        self._stop_event = Event()

    @cached_property
    def env(self) -> EnvTestEnvironment:
        """Return the underlying golang EnvTestEnvironment object. Must be called
        after starting the server

        Returns:
            EnvTestEnvironment: The underlying golang object for this instance
        """
        if not self._env:
            raise RuntimeError("Must start environment before accessing it")
        return self._env

    def start(self, auto_setup: bool = True) -> BytesIO:
        """Construct and start the EnvTestEnvironment object. Can optionally run setupenv-test
        to download the correct kube bins.

        Args:
            auto_setup (bool, optional): If setup-envtest should be ran to fetch the local
                path. Defaults to True.

        Returns:
            BytesIO: A raw kube_config for interacting with the cluster.
        """
        path = None
        if auto_setup:
            result = internal_run_setup_envtest_command("use", "-p", "path")
            result.log_and_raise(self._log, raise_exc=True)
            path = result.stdout

        if path:
            self._env = NewEnvTestEnvironmentWithPath(path)
        else:
            self._env = NewEnvTestEnvironment()

        self.env.Start()
        atexit.register(self.stop)

        return self.config()

    def stop(self):
        """Stop the envtest server if it hasn't already been stopped"""
        if self._stop_event.is_set():
            return

        self.env.Stop()
        self._stop_event.set()

    def config(self) -> BytesIO:
        """Fetch a kubeconfig object for the given EnvTest server. Must be called
        after start()

        Returns:
            BytesIO: The raw byte stream of the kube object
        """
        return BytesIO(bytes(self.env.GetKubeConfig()))

    def client(self) -> DynamicClient:
        """Fetch a dynamic client object for interacting with the EnvTest server. Must be called
        after start()

        Returns:
            DynamicClient: A client to interact with the server
        """
        return DynamicClient(self.config())



# ***************************************************************** #
# (C) Copyright IBM Corporation 2024.                               #
#                                                                   #
# The source code for this program is not published or otherwise    #
# divested of its trade secrets, irrespective of what has been      #
# deposited with the U.S. Copyright Office.                         #
# ***************************************************************** #
# SPDX-License-Identifier: Apache-2.0
"""This module contains a simple pythonic wrapper around the k8s.io/client-go/dynamic
DynamicClient.
"""

from kubernetes_lite.client.client import DynamicClient
from kubernetes_lite.client.resource import DynamicResource
from kubernetes_lite.client.sanitization import SourceObjectType
from kubernetes_lite.client.watch import Watch

__all__ = ["DynamicClient", "DynamicResource", "SourceObjectType", "Watch"]



# ***************************************************************** #
# (C) Copyright IBM Corporation 2024.                               #
#                                                                   #
# The source code for this program is not published or otherwise    #
# divested of its trade secrets, irrespective of what has been      #
# deposited with the U.S. Copyright Office.                         #
# ***************************************************************** #
# SPDX-License-Identifier: Apache-2.0
"""This module aids in comparing the performance and memory usage of
different kubernetes clients
"""



# ***************************************************************** #
# (C) Copyright IBM Corporation 2024.                               #
#                                                                   #
# The source code for this program is not published or otherwise    #
# divested of its trade secrets, irrespective of what has been      #
# deposited with the U.S. Copyright Office.                         #
# ***************************************************************** #
# SPDX-License-Identifier: Apache-2.0
"""This module contains useful testing utils that can't be stored in the
test or framework modules due to circular imports
"""

import statistics
import sys
import time
from contextlib import contextmanager


def unimport_module(module_name: str):
    """Helper to un-import the given module and its parents/siblings/
    children

    Arguments:
        module_name (str): The module name to unimport
    """
    if module_name in sys.modules:
        sys.modules.pop(module_name)

    # UnImport the module and any parent/sibling/child modules so
    # controller can be reimported from the most recent sys path
    module_parts = module_name.split(".")
    for i in range(1, len(module_parts)):
        parent_module = ".".join(module_parts[:-i])
        if parent_module in sys.modules:
            sys.modules.pop(parent_module, None)

    for child_module in [mod_name for mod_name in sys.modules if mod_name.startswith(f"{module_parts[0]}.")]:
        sys.modules.pop(child_module, None)


def calculate_timing_info(timing_list: list[float]) -> tuple[float, float]:
    """Calculate the average and stddiv for a list of floats

    Args:
        timing_list (list[float]): List of floats to calculate info about

    Returns:
        tuple[float, float]: The average and stddiv for the list
    """
    if len(timing_list) < 1:
        raise ValueError("Can't calculate timing information with empty list")

    stddev = 0
    average = timing_list[0] * 1000
    if len(timing_list) > 1:
        stddev = statistics.stdev(timing_list) * 1000
        average = statistics.mean(timing_list) * 1000
    return average, stddev


@contextmanager
def time_for_list(result_list: list[float]):
    """Context manager to aid in timing a function and appending it to an existing list
    of float

    Args:
        result_list (list[float]): the list to append timing information to
    """
    start_time = time.perf_counter()
    yield
    end_time = time.perf_counter()
    result_list.append(end_time - start_time)



# ***************************************************************** #
# (C) Copyright IBM Corporation 2024.                               #
#                                                                   #
# The source code for this program is not published or otherwise    #
# divested of its trade secrets, irrespective of what has been      #
# deposited with the U.S. Copyright Office.                         #
# ***************************************************************** #
# SPDX-License-Identifier: Apache-2.0
"""This module contains the implementation of ClientInterface using the local kubernetes-lite
package
"""

import os
from pathlib import Path
from typing import Any

from performance.interfaces.base import ClientInterface


class KubernetesLiteInterface(ClientInterface):
    """KubernetesLiteInterface is the implementation of ClientInterface using the local kubernetes-lite
    library
    """

    @staticmethod
    def import_modules() -> list[str]:
        """Import the kubernetes_lite dynamic client since that's all thats required"""
        from kubernetes_lite.client import DynamicClient  # noqa: F401, PLC0415

        return ["kubernetes_lite"]

    @staticmethod
    def create_client() -> Any:
        """Create the client using the local kube_config"""
        from kubernetes_lite.client import DynamicClient  # noqa: PLC0415

        config = None
        # Manually read in KUBECONFIG. This is required in testing because the golang runtime copies the environment
        # before we update the os.environ variable
        if os.environ.get("KUBECONFIG"):
            config = Path(os.environ.get("KUBECONFIG")).read_bytes()
        return DynamicClient(config=config)

    @staticmethod
    def create_resource_client(client: Any, kind: str, api_version: str) -> Any:
        """Create the dynamic resource client with the provided kind/apiVerison"""
        return client.resource(api_version, kind)

    @staticmethod
    def create(resource_client: Any, data: Any) -> Any:
        """Create an object in the cluster"""
        return resource_client.create(data)

    @staticmethod
    def apply(resource_client: Any, data: Any) -> Any:
        """Apply an object in the cluster"""
        return resource_client.apply(data, field_manager="python")

    @staticmethod
    def list(resource_client: Any, namespace: str) -> Any:
        """List all objects in the namespace"""
        return resource_client.list(namespace)

    @staticmethod
    def get(resource_client: Any, name: str, namespace: str) -> Any:
        """Get an single object in the cluster"""
        return resource_client.get(name=name, namespace=namespace)

    @staticmethod
    def delete(resource_client: Any, name: str, namespace: str):
        """Delete a single object in the cluster"""
        resource_client.delete(name=name, namespace=namespace)



# ***************************************************************** #
# (C) Copyright IBM Corporation 2024.                               #
#                                                                   #
# The source code for this program is not published or otherwise    #
# divested of its trade secrets, irrespective of what has been      #
# deposited with the U.S. Copyright Office.                         #
# ***************************************************************** #
# SPDX-License-Identifier: Apache-2.0
# Assisted by watsonx Code Assistant
"""This example program demonstrates the fundamental operations for
managing on Deployment resources, such as Create, List, Update and Delete
using kubernetes_lites dynamic package.
"""

import argparse
import time
from collections.abc import Callable
from pathlib import Path

from kubernetes_lite.client import DynamicClient
from kubernetes_lite.errors import ConflictError


def exponential_retry(
    func: Callable, caught_exception: Exception, max_attempts: int, initial_delay: float, max_delay: float
):
    """Helper function to exponentially retry a function call for certain exceptions. This
    isn't explicitly needed for the example but it aligns with the client-go example

    Args:
        func (Callable): The function to call/retry
        caught_exception (Exception): The exception to catch for retries
        max_attempts (int): How many times to retry
        initial_delay (float): What the initial exponential delay is
        max_delay (float): Maximum time delay between retries
    """
    attempts = 0
    delay = initial_delay
    last_exception = None
    while attempts < max_attempts:
        try:
            func()
            return
        except caught_exception as exc:
            last_exception = exc

        time.sleep(delay)
        delay = min(delay * 2, max_delay)
        attempts += 1
    raise last_exception


def main(kube_config: Path | None = None, namespace: str | None = "default"):
    """_summary_

    Args:
        kube_config (Path | None, optional): _description_. Defaults to None.
        namespace (str | None, optional): _description_. Defaults to "default".
    """
    client = DynamicClient(kube_config.read_bytes()) if kube_config else DynamicClient()

    deploy_resource = client.resource("apps/v1", "deployments")

    deployment_obj = {
        "apiVersion": "apps/v1",
        "kind": "Deployment",
        "metadata": {
            "name": "demo-deployment",
        },
        "spec": {
            "replicas": 2,
            "selector": {
                "matchLabels": {
                    "app": "demo",
                },
            },
            "template": {
                "metadata": {
                    "labels": {
                        "app": "demo",
                    },
                },
                "spec": {
                    "containers": [
                        {
                            "name": "web",
                            "image": "nginx:1.12",
                            "ports": [
                                {
                                    "name": "http",
                                    "protocol": "TCP",
                                    "containerPort": 80,
                                },
                            ],
                        }
                    ],
                },
            },
        },
    }

    print("Creating deployment...")
    result = deploy_resource.create(deployment_obj, namespace=namespace)
    print(f'Created deployment "{result["metadata"]["name"]}"')

    input("-> Press Return key to continue.")

    print("Updating deployment...")

    def update_deployment():
        current_deploy_obj = deploy_resource.get(name="demo-deployment", namespace=namespace)
        current_deploy_obj["spec"]["replicas"] = 1

        current_deploy_obj["spec"]["template"]["spec"]["containers"][0]["image"] = "nginx:1.13"

        deploy_resource.update(current_deploy_obj)

    exponential_retry(update_deployment, caught_exception=ConflictError, max_attempts=5, initial_delay=1, max_delay=10)
    print("Updated deployment...")
    input("-> Press Return key to continue.")

    print(f'Listing deployments in namespace "{namespace}":')
    resources = deploy_resource.list(namespace=namespace)
    for resource in resources["items"]:
        name = resource["metadata"]["name"]
        replicas = resource["spec"]["replicas"]
        print(f" * {name} ({replicas} replicas)")

    input("-> Press Return key to continue.")

    print("Deleting deployment...")
    deploy_resource.delete(
        name="demo-deployment",
        namespace=namespace,
        propagation_policy="Foreground",
    )
    print("Deleted deployment.")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-kubeconfig", help="Path to the kubeconfig file")
    parser.add_argument("-namespace", help="Namespace to use", default="default")
    args = parser.parse_args()

    kubeconfig = None
    if args.kubeconfig:
        kubeconfig = Path(args.kubeconfig)
        if not kubeconfig.is_file():
            raise argparse.ArgumentTypeError(f"kubeconfig: {kubeconfig} either does not exist or is not a file")

    main(kubeconfig, args.namespace)



# ***************************************************************** #
# (C) Copyright IBM Corporation 2024.                               #
#                                                                   #
# The source code for this program is not published or otherwise    #
# divested of its trade secrets, irrespective of what has been      #
# deposited with the U.S. Copyright Office.                         #
# ***************************************************************** #
# SPDX-License-Identifier: Apache-2.0
"""This module contains the implementation of ClientInterface using kr8s"""

from typing import Any

from performance.interfaces.base import ClientInterface


class Kr8sInterface(ClientInterface):
    """Kr8sInterface is a client interface implementation built using the k8rs
    library provided at https://github.com/kr8s-org/kr8s
    """

    @staticmethod
    def import_modules() -> list[str]:
        """Import the kr8s module"""
        import kr8s  # noqa: F401, PLC0415

        return ["kr8s"]

    @staticmethod
    def create_client() -> Any:
        """The kr8s module acts as the client since its under the cover"""
        import kr8s  # noqa: PLC0415

        return kr8s

    @staticmethod
    def create_resource_client(client: Any, kind: str, api_version: str) -> Any:
        """Create resource client constructs a new kr8s class for the provided kind/apiVersion"""
        from kr8s.objects import new_class  # noqa: PLC0415

        namespaced = True
        if kind == "CustomResourceDefinition":
            namespaced = False
        return new_class(
            kind=kind,
            version=api_version,
            namespaced=namespaced,
        )

    @staticmethod
    def create(resource_client: Any, data: Any) -> Any:
        """Create an object using kr8s's create api"""
        res_obj = resource_client(resource=data)
        res_obj.create()
        res_obj.refresh()
        return res_obj.to_dict()

    @staticmethod
    def apply(resource_client: Any, data: Any) -> Any:
        """Apply an object using either patch/create for kr8s"""
        res_obj = resource_client(resource=data)
        if res_obj.exists():
            res_obj.patch(data)
        else:
            res_obj.create()
        res_obj.refresh()
        return res_obj.to_dict()

    @staticmethod
    def list(resource_client: Any, namespace: str) -> Any:
        """List all objects using the kr8s iterator"""
        return [res.to_dict() for res in resource_client.list(namespace=namespace)]

    @staticmethod
    def get(resource_client: Any, name: str, namespace: str) -> Any:
        """Get the object from the cluster using kr8s's get interface"""
        return resource_client.get(name=name, namespace=namespace).to_dict()

    @staticmethod
    def delete(resource_client: Any, name: str, namespace: str):
        """Delete the object from the cluster after refreshing it"""
        res = resource_client(name, namespace)
        res.delete()



# ***************************************************************** #
# (C) Copyright IBM Corporation 2024.                               #
#                                                                   #
# The source code for this program is not published or otherwise    #
# divested of its trade secrets, irrespective of what has been      #
# deposited with the U.S. Copyright Office.                         #
# ***************************************************************** #
# SPDX-License-Identifier: Apache-2.0
"""This module contains the interface for iterating over events from the
cluster
"""

from kubernetes_lite.serialization import deserialize_result
from kubernetes_lite.wrapper.client import WrappedWatchInterface


class Watch:
    """Watch is an iterator that yields events streamed from
    the cluster
    """

    interface: WrappedWatchInterface | None = None

    def __init__(self, interface: WrappedWatchInterface | None = None):
        """Initialize the Watch instance with a given wrapped interface

        Args:
            interface (WrappedWatchInterface | None, optional): The interface to yield events from. Defaults to None.
        """
        self.interface = interface

    def stop(self):
        """Stop streaming events from the cluster"""
        if not self.interface:
            raise RuntimeError("Interface must be set before calling stop/next")

        self.interface.Stop()

    def __iter__(self) -> "Watch":
        """Complete implementation of a python iterator

        Returns:
            Watch: Itself to iterate
        """
        return self

    def __next__(self) -> dict:
        """Return the next event from the cluster. These events are dictionaries in the
        format of json encoded apimachinery Events. You can find more information here:
        https://pkg.go.dev/k8s.io/apimachinery/pkg/watch#Event

        Returns:
            dict: The event data
        """
        if not self.interface:
            raise RuntimeError("Interface must be set before calling stop/next")

        # Try to get an event from the underlying interface
        try:
            event_data = self.interface.Next()
        except RuntimeError as exc:
            if "stop iteration" in str(exc):
                raise StopIteration from None

        # Return the deserialized result
        return deserialize_result(bytes(event_data))



# ***************************************************************** #
# (C) Copyright IBM Corporation 2024.                               #
#                                                                   #
# The source code for this program is not published or otherwise    #
# divested of its trade secrets, irrespective of what has been      #
# deposited with the U.S. Copyright Office.                         #
# ***************************************************************** #
# SPDX-License-Identifier: Apache-2.0
"""This module contains helper dataclasses/functions for parsing exit data
information from setupenv-test
"""

import re
from dataclasses import dataclass

# Regex for parsing the exit code and text from a string. Here is an expected text this regex
# can parse "some error message (exit code 2)"
EXIT_CODE_REGEX = re.compile(r"(?P<text>.*)\(exit code (?P<code>[0-9])\).*")


@dataclass
class ExitData:
    """ExitData dataclass contains the exit code and message for a setup-envtest
    command run

    Attributes:
        code int: The exit code that was parsed
        message (str | None): An optional string message if one was extracted
    """

    code: int
    message: str | None

    @classmethod
    def from_string(cls, error: str) -> "ExitData|None":
        """Attempt to parse a error string into a ExitData dataclass

        Returns:
            ExitData|None: The parsed ExitData or None if one couldn't
                be extracted
        """
        # Attempt to match the regex
        match = EXIT_CODE_REGEX.match(error)
        if not match:
            return

        matches = match.groupdict()
        exit_code = matches.get("code", None)
        message = matches.get("text", None).strip()
        if message:
            message = None

        if exit_code:
            try:
                exit_code = int(exit_code)
            except ValueError:
                return

        return cls(code=exit_code, message=message)



# ***************************************************************** #
# (C) Copyright IBM Corporation 2024.                               #
#                                                                   #
# The source code for this program is not published or otherwise    #
# divested of its trade secrets, irrespective of what has been      #
# deposited with the U.S. Copyright Office.                         #
# ***************************************************************** #
# SPDX-License-Identifier: Apache-2.0
"""This module contains the main implementation of the performance command
which includes formatting the output and running the test harness
"""

import csv
import os
import time
from enum import StrEnum
from io import BytesIO
from pathlib import Path
from tempfile import TemporaryDirectory

from performance.interfaces.kr8s_interface import Kr8sInterface
from performance.interfaces.kubernetes_interface import KubernetesInterface
from performance.interfaces.kubernetes_lite_interface import KubernetesLiteInterface
from performance.interfaces.run import TimingInfo, test_interface
from performance.memory import BackgroundMemoryTracker
from performance.utils import calculate_timing_info


class ClientOptions(StrEnum):
    """Enum of the currently supported clients that we can measure"""

    KUBERNETES = "kubernetes"
    KUBERNETES_LITE = "kubernetes_lite"
    KR8S = "kr8s"


# This dictionary maps the ClientOption from the CLI to the corresponding interface
CLIENT_OPTION_INTERFACE_MAP = {
    ClientOptions.KUBERNETES_LITE: KubernetesLiteInterface,
    ClientOptions.KUBERNETES: KubernetesInterface,
    ClientOptions.KR8S: Kr8sInterface,
}


def run_test(
    output: Path,
    clients: list[ClientOptions] | None = None,
    repetitions: int = 5,
    namespace: str = "default",
    memory_poll_time: float = 0.1,
):
    """Run the performance test for a given list of clients. This is the functional entrypoint
    for this module

    Args:
        output (Path): Output file to write results to
        clients (list[ClientOptions] | None, optional): List of clients to test. Defaults to None which
            means test all clients
        repetitions (int, optional): How many times the test should be ran. Defaults to 5.
        namespace (str, optional): Namespace to run test in. Defaults to "default" and should
            be unique if multiple performance tests are running at the same time
        memory_poll_time (float, optional): How often to poll the system memory.
            Defaults to 0.1.
    """
    # If no client was provided then test them all
    if not clients:
        clients = [client for client in ClientOptions]

    # Before doing anything start the memory tracker so we can track the initial
    # kubernetes_lite usage
    memory_tracker = BackgroundMemoryTracker(poll_time=memory_poll_time)
    memory_tracker.start()

    # Import env test and sleep to ensure the tracker captured some data
    time.sleep(1)
    from kubernetes_lite.envtest import EnvTest  # noqa: PLC0415

    time.sleep(1)

    # Get kubernetes_lite max mem
    kubernetes_lite_mem_usage = memory_tracker.get_max_mem()

    # Setup EnvTest server
    env_test = EnvTest()
    env_test.start()
    config = env_test.config()

    # Attempt to test the client interfaces
    try:
        # Write the envtest kube config to a temporary directory
        timing_outputs, max_mem = test_interfaces(clients, config, memory_tracker, repetitions, namespace)
    # Even if there was an error shutdown the tracker/envtest
    finally:
        memory_tracker.shutdown()
        memory_tracker.join()
        env_test.stop()

    # Update the max mem for kubernetes_lite to the value we gathered after importing EnvTest. EnvTest overall increases
    # reproducibility but it requires measuring max_mem separately
    if ClientOptions.KUBERNETES_LITE in max_mem:
        max_mem[ClientOptions.KUBERNETES_LITE] = kubernetes_lite_mem_usage
    if len(max_mem) == 1 and ClientOptions.KUBERNETES_LITE not in max_mem:
        max_mem_key = next(iter(max_mem.keys()))
        max_mem[max_mem_key] -= kubernetes_lite_mem_usage

    # After running the tests generate the CSV output
    generate_result_csv(output, timing_outputs, max_mem)


def test_interfaces(
    clients: list[ClientOptions],
    config: BytesIO,
    memory_tracker: BackgroundMemoryTracker,
    repetitions: int,
    namespace: str,
) -> tuple[dict[ClientOptions, TimingInfo], dict[ClientOptions, float]]:
    """Test and time the various selected interfaces. This is split out to simplify the
    run_test function

    Args:
        clients (list[ClientOptions]): Clients to test
        config (BytesIO): KubeConfig to use during testing
        memory_tracker (BackgroundMemoryTracker): The memory tracker running in the background
        repetitions (int): How many repetitions to perform
        namespace (str): The namespace to run the test against

    Returns:
        tuple[dict[ClientOptions, TimingInfo], dict[ClientOptions, float]]: tuple of timing information and max_memory
    """
    # Setup a tempdir with the kube config. THis is required for some of the simpler clients
    with TemporaryDirectory() as temp_dir:
        temp_path = Path(temp_dir)
        config_path = temp_path / "config"
        config_path.write_bytes(config.getvalue())
        os.environ["KUBECONFIG"] = str(config_path)

        # For each client run the tests
        timing_outputs: dict[ClientOptions, TimingInfo] = {}
        max_mem = {}
        for client in clients:
            # Fetch and construct the interface
            interface_type = CLIENT_OPTION_INTERFACE_MAP.get(client)
            interface = interface_type()

            # Reset the memory tracker before running the test
            memory_tracker.reset()
            timing_outputs[client] = test_interface(interface, repetitions, namespace)

            # Get max mem after the test
            max_mem[client] = memory_tracker.get_max_mem()

    return timing_outputs, max_mem


def generate_result_csv(output: Path, timing_outputs: dict[ClientOptions, TimingInfo], max_mem: dict[str, float]):
    """Generate a CSV document from the provided timing and memory outputs

    Args:
        output (Path): The path to write the CSV to
        timing_outputs (dict[ClientOptions, TimingInfo]): Dictionary of client types and their timing information
        max_mem (dict[str, float]): Each client's max RSS memory usage
    """
    output_csv_rows = [
        [
            "TYPE",
            "FILE",
            "METHOD",
            "AVERAGE TIME (ms)",
            "STD DIV (ms)",
        ]
    ]
    for output_type, timing_info in timing_outputs.items():
        output_csv_rows.extend(
            [
                [output_type.value, "N/A", "client_create", *calculate_timing_info(timing_info.client_create_times)],
                [
                    output_type.value,
                    "N/A",
                    "resource_client_create_times",
                    *calculate_timing_info(timing_info.resource_client_create_times),
                ],
                [output_type.value, "N/A", "import_times", *calculate_timing_info(timing_info.import_times)],
            ]
        )

        for file_name, resource_timing_info in timing_info.resource_timing_info.items():
            output_csv_rows.extend(
                [
                    [output_type.value, file_name, "apply", *calculate_timing_info(resource_timing_info.apply_times)],
                    [output_type.value, file_name, "delete", *calculate_timing_info(resource_timing_info.delete_times)],
                    [output_type.value, file_name, "create", *calculate_timing_info(resource_timing_info.create_times)],
                    [output_type.value, file_name, "list", *calculate_timing_info(resource_timing_info.list_times)],
                    [output_type.value, file_name, "get", *calculate_timing_info(resource_timing_info.get_times)],
                ]
            )

    # Add two blank lines before memory summary
    output_csv_rows.append([])
    output_csv_rows.append([])

    output_csv_rows.append(
        [
            "TYPE",
            "MAX_RSS_INCREASE",
        ]
    )
    for output_type, max_rss in max_mem.items():
        output_row = [output_type.value, max_rss]
        output_csv_rows.append(output_row)

    with output.open("w") as output_file:
        writer = csv.writer(output_file)
        writer.writerows(output_csv_rows)



# ***************************************************************** #
# (C) Copyright IBM Corporation 2024.                               #
#                                                                   #
# The source code for this program is not published or otherwise    #
# divested of its trade secrets, irrespective of what has been      #
# deposited with the U.S. Copyright Office.                         #
# ***************************************************************** #
# SPDX-License-Identifier: Apache-2.0
"""This module contains the abstract base class for all client interfaces"""

import abc
from typing import Any


class ClientInterface(abc.ABC):
    """ClientInterface is the general base class for clients. It contains the description and signature
    for all of the expected methods
    """

    @staticmethod
    @abc.abstractmethod
    def import_modules() -> list[str]:
        """Just import the modules required to use the client with no other tasks.

        Returns:
            list[str]: List of modules that were imported so they can be unimported before being reran
        """

    @staticmethod
    @abc.abstractmethod
    def create_client() -> Any:
        """Create a client object assuming that the KUBE_CONFIG variable is properly set

        Returns:
            Any: Functioning client which can be passed into create_resource_client
        """

    @staticmethod
    @abc.abstractmethod
    def create_resource_client(client: Any, kind: str, api_version: str) -> Any:
        """Create a resource specific client

        Args:
            client (Any): The client to use for connections
            kind (str): The  kind of the resource client to create
            api_version (str): The apiVersion of the resource client to create

        Returns:
            Any: The constructed resource client
        """

    @staticmethod
    @abc.abstractmethod
    def create(resource_client: Any, data: Any) -> Any:
        """Create an object with a given resource_client

        Args:
            resource_client (Any): The resource_client to use for operations
            data (Any): The data to create

        Returns:
            Any: The current object's cluster state
        """

    @staticmethod
    @abc.abstractmethod
    def apply(resource_client: Any, data: Any) -> Any:
        """Apply an object to the cluster with a given resource_client

        Args:
            resource_client (Any): The resource_client to use for operations
            data (Any): The data to create

        Returns:
            Any: The current object's cluster state
        """

    @staticmethod
    @abc.abstractmethod
    def list(resource_client: Any, namespace: str | None) -> Any:
        """List all objects in the cluster for a given namespace/resource_client

        Args:
            resource_client (Any): The resource_client to use for operations
            namespace (str): The namespace to list resources in

        Returns:
            Any: List of current resources
        """

    @staticmethod
    @abc.abstractmethod
    def get(resource_client: Any, name: str, namespace: str | None) -> Any:
        """Get an object from the cluster

        Args:
            resource_client (Any): THe resource_client to use for operations
            name (str): The name of the object to fetch
            namespace (str | None): The namespace the object lives in

        Returns:
            Any: The object from the cluster as a dictionary
        """

    @staticmethod
    @abc.abstractmethod
    def delete(resource_client: Any, name: str, namespace: str):
        """Delete an object from the cluster

        Args:
            resource_client (Any): THe resource_client to use for operations
            name (str): The name of the object to delete
            namespace (str | None): The namespace the object lives in
        """



# ***************************************************************** #
# (C) Copyright IBM Corporation 2024.                               #
#                                                                   #
# The source code for this program is not published or otherwise    #
# divested of its trade secrets, irrespective of what has been      #
# deposited with the U.S. Copyright Office.                         #
# ***************************************************************** #
# SPDX-License-Identifier: Apache-2.0
"""This module tests that setup-envtest functions like the go version"""

import subprocess
from pathlib import Path


def test_help():
    """Test that help text is displayed if --help is passed in"""
    proc = subprocess.run(
        ["python3", "-m", "kubernetes_lite.setup_envtest", "--help"], capture_output=True, check=False
    )
    assert proc.returncode == 2
    assert b"Usage: python3 -m kubernetes_lite.setup_envtest" in proc.stderr


def test_use():
    """Test that we can get the path from envtest"""
    proc = subprocess.run(
        ["python3", "-m", "kubernetes_lite.setup_envtest", "use", "-p", "path"], capture_output=True, check=False
    )
    assert proc.returncode == 0
    file_path = Path(proc.stdout.decode("utf-8"))
    assert file_path.exists()
    assert file_path.is_dir()



# ***************************************************************** #
# (C) Copyright IBM Corporation 2024.                               #
#                                                                   #
# The source code for this program is not published or otherwise    #
# divested of its trade secrets, irrespective of what has been      #
# deposited with the U.S. Copyright Office.                         #
# ***************************************************************** #
# SPDX-License-Identifier: Apache-2.0
"""This module contains tests for the various resource operations"""

# Assisted by watsonx Code Assistant
import copy

from kubernetes_lite.client import DynamicResource
from kubernetes_lite.errors import ConflictError, NotFoundError
from tests.utils import (
    check_object_status,
    generate_test_object,
    validate_result,
)

import pytest


def test_create(resource_namespace: tuple[DynamicResource, str]):
    """Test that the client can create objects"""
    deployment_res, namespace = resource_namespace
    obj = generate_test_object(namespace)

    deployment_res.create(obj)

    check_object_status(deployment_res, expected_obj=obj)


def test_update(resource_namespace: tuple[DynamicResource, str]):
    """Test that the client can update objects"""
    deployment_res, namespace = resource_namespace
    org_obj = generate_test_object(namespace)

    # Ensure we get a NotFoundError if we update a nonexistent resource
    with pytest.raises(NotFoundError):
        deployment_res.update(org_obj)

    # Create initial resource before updating
    deployment_res.create(org_obj)

    updated_obj = copy.deepcopy(org_obj)
    updated_obj["metadata"]["labels"]["new-label"] = "test"

    deployment_res.update(updated_obj)

    check_object_status(deployment_res, expected_obj=updated_obj)


def test_update_status(resource_namespace: tuple[DynamicResource, str]):
    """Test that the client can update the status of objects"""
    deployment_res, namespace = resource_namespace
    obj = generate_test_object(namespace)

    current_obj = deployment_res.create(obj)
    assert not current_obj["status"].get("availableReplicas")

    current_obj["status"]["replicas"] = 1
    deployment_res.update_status(current_obj)

    check_object_status(deployment_res, expected_obj=current_obj)


def test_patch(resource_namespace: tuple[DynamicResource, str]):
    """Test that the client can patch objects with merge"""
    deployment_res, namespace = resource_namespace
    obj = generate_test_object(namespace)

    current_obj = deployment_res.create(obj)
    assert current_obj["spec"].get("replicas") == 3

    patch = {"spec": {"replicas": 1}}
    deployment_res.patch(
        patch_type="application/strategic-merge-patch+json",
        patch_data=patch,
        name=current_obj.get("metadata").get("name"),
        namespace=namespace,
    )

    current_obj["spec"]["replicas"] = 1
    check_object_status(deployment_res, expected_obj=current_obj)


def test_apply(resource_namespace: tuple[DynamicResource, str]):
    """Test that the client can apply objects"""
    deploy_res, namespace = resource_namespace
    obj = generate_test_object(namespace)

    check_object_status(deploy_res, namespace, exists=False)

    deploy_res.apply(obj, field_manager="test")
    check_object_status(deploy_res, expected_obj=obj)

    obj["spec"]["replicas"] = 2
    deploy_res.apply(obj, field_manager="test")
    check_object_status(deploy_res, expected_obj=obj)

    obj["spec"]["replicas"] = 3
    with pytest.raises(ConflictError):
        deploy_res.apply(obj, field_manager="other")
    deploy_res.apply(obj, field_manager="other", force=True)
    check_object_status(deploy_res, expected_obj=obj)


def test_apply_status(resource_namespace: tuple[DynamicResource, str]):
    """Test that the client can apply the status of objects"""
    deployment_res, namespace = resource_namespace
    obj = generate_test_object(namespace)

    current_obj = deployment_res.create(obj)
    assert not current_obj["status"].get("replicas")

    obj["status"] = {}
    obj["status"]["replicas"] = 1
    deployment_res.apply_status(obj, field_manager="test", force=True)

    check_object_status(deployment_res, expected_obj=obj)


def test_delete(resource_namespace: tuple[DynamicResource, str]):
    """Test that the client can delete objects"""
    deployment_resource, namespace = resource_namespace
    obj = generate_test_object(namespace)
    obj_metadata = obj.get("metadata")

    with pytest.raises(NotFoundError):
        deployment_resource.delete(obj_metadata.get("name"), obj_metadata.get("namespace"))

    # Create before attempting to delete for real
    deployment_resource.create(obj)

    deployment_resource.delete(obj_metadata.get("name"), obj_metadata.get("namespace"))
    check_object_status(deployment_resource, namespace, exists=False)


def test_delete_collection(resource_namespace: tuple[DynamicResource, str]):
    """Test that the client can delete a list of objects"""
    deployment_resource, namespace = resource_namespace

    # Assert that we can delete a collection with no objects
    deployment_resource.delete_collection(namespace=namespace)

    # Create before attempting to delete for real
    obj_1 = generate_test_object(namespace, name="test-1")
    obj_2 = generate_test_object(namespace, name="test-2")
    deployment_resource.create(obj_1)
    deployment_resource.create(obj_2)

    check_object_status(deployment_resource, namespace, name="test-1", exists=True)
    check_object_status(deployment_resource, namespace, name="test-2", exists=True)

    deployment_resource.delete_collection(namespace=namespace)
    check_object_status(deployment_resource, namespace, name="test-1", exists=False)
    check_object_status(deployment_resource, namespace, name="test-2", exists=False)


def test_list(resource_namespace: tuple[DynamicResource, str]):
    """Test that the client can list objects"""
    deployment_resource, namespace = resource_namespace

    # With no objects assert list returns empty
    empty_list_resp = deployment_resource.list(namespace=namespace)
    assert empty_list_resp
    assert empty_list_resp.get("items") == []

    obj_1 = generate_test_object(namespace, name="test-1")
    obj_2 = generate_test_object(namespace, name="test-2")
    deployment_resource.create(obj_1)
    deployment_resource.create(obj_2)

    list_resp = deployment_resource.list(namespace=namespace)
    assert list_resp
    assert len(list_resp.get("items")) == 2
    validate_result(list_resp.get("items")[0], obj_1)
    validate_result(list_resp.get("items")[1], obj_2)



# ***************************************************************** #
# (C) Copyright IBM Corporation 2024.                               #
#                                                                   #
# The source code for this program is not published or otherwise    #
# divested of its trade secrets, irrespective of what has been      #
# deposited with the U.S. Copyright Office.                         #
# ***************************************************************** #
# SPDX-License-Identifier: Apache-2.0
"""This script is an example of how to use envtest with the standard
kubernetes package
"""

from kubernetes_lite.envtest import EnvTest
from kubernetes_lite.envtest.pytest import session_kubernetes_env  # noqa: F401

from kubernetes import client, config
from kubernetes.dynamic import DynamicClient


def test_kubernetes_core_api(session_kubernetes_env: EnvTest):  # noqa: F811
    """Test that kubernetes core api works with envtest"""
    config.load_config(config_file=session_kubernetes_env.config())

    v1 = client.CoreV1Api()
    print("Listing pods with their IPs:")
    ret = v1.list_pod_for_all_namespaces(watch=False)
    for i in ret.items:
        print(f"{i.status.pod_ip}\t{i.metadata.namespace}\t{i.metadata.name}")

    assert ret


def test_kubernetes_dynamic(session_kubernetes_env: EnvTest):  # noqa: F811
    """Test that kubernetes dynamic client works with envtest"""
    kube_config = config.new_client_from_config(config_file=session_kubernetes_env.config())
    kube_client = DynamicClient(kube_config)
    assert kube_client

    deploy_res = kube_client.resources.get(kind="Deployment", api_version="apps/v1")
    ret = deploy_res.get(name=None, namespace=None)
    for i in ret.items:
        print(f"{i.status.pod_ip}\t{i.metadata.namespace}\t{i.metadata.name}")
    assert ret




# python wrapper for package github.ibm.com/Michael-Honaker/kubernetes-lite/kubernetes_lite/go_wrapper/pkg/client within overall package wrapper
# This is what you import to use the package.
# File is generated by gopy. Do not edit.
# gopy pkg -no-make -author=Michael Honaker -email=michael.honaker@ibm.com -name=wrapper github.ibm.com/Michael-Honaker/kubernetes-lite/kubernetes_lite/go_wrapper/pkg/client github.ibm.com/Michael-Honaker/kubernetes-lite/kubernetes_lite/go_wrapper/pkg/envtest/setup github.ibm.com/Michael-Honaker/kubernetes-lite/kubernetes_lite/go_wrapper/pkg/envtest/server

# the following is required to enable dlopen to open the _go.so file
import os,sys,inspect,collections
try:
	import collections.abc as _collections_abc
except ImportError:
	_collections_abc = collections

cwd = os.getcwd()
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
os.chdir(currentdir)
from . import _wrapper
from . import go

os.chdir(cwd)

# to use this code in your end-user python file, import it as follows:
# from wrapper import client
# and then refer to everything using client. prefix
# packages imported by this package listed below:




# ---- Types ---


#---- Enums from Go (collections of consts with same type) ---


#---- Constants from Go: Python can only ask that you please don't change these! ---


# ---- Global Variables: can only use functions to access ---
def EmptyEvent():
	"""
	EmptyEvent Gets Go Variable: client.EmptyEvent
	
	"""
	return go.watch_Event(handle=_wrapper.client_EmptyEvent())

def Set_EmptyEvent(value):
	"""
	Set_EmptyEvent Sets Go Variable: client.EmptyEvent
	
	"""
	if isinstance(value, go.GoClass):
		_wrapper.client_Set_EmptyEvent(value.handle)
	else:
		_wrapper.client_Set_EmptyEvent(value)



# ---- Interfaces ---

# Python type for interface client.WrappedDynamicClient
class WrappedDynamicClient(go.GoClass):
	"""Simple dynamic client interface that allows users to create a Resource Interface\n"""
	def __init__(self, *args, **kwargs):
		"""
		handle=A Go-side object is always initialized with an explicit handle=arg
		"""
		if len(kwargs) == 1 and 'handle' in kwargs:
			self.handle = kwargs['handle']
			_wrapper.IncRef(self.handle)
		elif len(args) == 1 and isinstance(args[0], go.GoClass):
			self.handle = args[0].handle
			_wrapper.IncRef(self.handle)
		else:
			self.handle = 0
	def Resource(self, apiVersion, kind):
		"""Resource(str apiVersion, str kind) object, str"""
		return WrappedNamespaceableResourceInterface(handle=_wrapper.client_WrappedDynamicClient_Resource(self.handle, apiVersion, kind))

# Python type for interface client.WrappedNamespaceableResourceInterface
class WrappedNamespaceableResourceInterface(go.GoClass):
	"""WrappedNamespaceableResourceInterface is a wrapper around the ResourceInterface for interacting with\ncluster wide resources. This interface can also be namespace scoped with the Namespace method.\n"""
	def __init__(self, *args, **kwargs):
		"""
		handle=A Go-side object is always initialized with an explicit handle=arg
		"""
		if len(kwargs) == 1 and 'handle' in kwargs:
			self.handle = kwargs['handle']
			_wrapper.IncRef(self.handle)
		elif len(args) == 1 and isinstance(args[0], go.GoClass):
			self.handle = args[0].handle
			_wrapper.IncRef(self.handle)
		else:
			self.handle = 0
	def Apply(self, name, obj, opts, *args):
		"""Apply(str name, []int obj, []int opts, []str subresources) []int, str"""
		subresources = go.Slice_string(args)
		return go.Slice_byte(handle=_wrapper.client_WrappedNamespaceableResourceInterface_Apply(self.handle, name, obj.handle, opts.handle, subresources.handle))
	def ApplyStatus(self, name, obj, opts):
		"""ApplyStatus(str name, []int obj, []int opts) []int, str"""
		return go.Slice_byte(handle=_wrapper.client_WrappedNamespaceableResourceInterface_ApplyStatus(self.handle, name, obj.handle, opts.handle))
	def Create(self, obj, opts, *args):
		"""Create([]int obj, []int opts, []str subresources) []int, str"""
		subresources = go.Slice_string(args)
		return go.Slice_byte(handle=_wrapper.client_WrappedNamespaceableResourceInterface_Create(self.handle, obj.handle, opts.handle, subresources.handle))
	def Delete(self, name, opts, *args):
		"""Delete(str name, []int opts, []str subresources) str"""
		subresources = go.Slice_string(args)
		return _wrapper.client_WrappedNamespaceableResourceInterface_Delete(self.handle, name, opts.handle, subresources.handle)
	def DeleteCollection(self, opts, listOptions):
		"""DeleteCollection([]int opts, []int listOptions) str"""
		return _wrapper.client_WrappedNamespaceableResourceInterface_DeleteCollection(self.handle, opts.handle, listOptions.handle)
	def Get(self, name, opts, *args):
		"""Get(str name, []int opts, []str subresources) []int, str"""
		subresources = go.Slice_string(args)
		return go.Slice_byte(handle=_wrapper.client_WrappedNamespaceableResourceInterface_Get(self.handle, name, opts.handle, subresources.handle))
	def List(self, opts):
		"""List([]int opts) []int, str"""
		return go.Slice_byte(handle=_wrapper.client_WrappedNamespaceableResourceInterface_List(self.handle, opts.handle))
	def Namespace(self, arg_0):
		"""Namespace(str) object"""
		return WrappedResourceInterface(handle=_wrapper.client_WrappedNamespaceableResourceInterface_Namespace(self.handle, arg_0))
	def Patch(self, name, pt, data, opts, *args):
		"""Patch(str name, str pt, []int data, []int opts, []str subresources) []int, str"""
		subresources = go.Slice_string(args)
		return go.Slice_byte(handle=_wrapper.client_WrappedNamespaceableResourceInterface_Patch(self.handle, name, pt, data.handle, opts.handle, subresources.handle))
	def Update(self, obj, opts, *args):
		"""Update([]int obj, []int opts, []str subresources) []int, str"""
		subresources = go.Slice_string(args)
		return go.Slice_byte(handle=_wrapper.client_WrappedNamespaceableResourceInterface_Update(self.handle, obj.handle, opts.handle, subresources.handle))
	def UpdateStatus(self, obj, opts):
		"""UpdateStatus([]int obj, []int opts) []int, str"""
		return go.Slice_byte(handle=_wrapper.client_WrappedNamespaceableResourceInterface_UpdateStatus(self.handle, obj.handle, opts.handle))
	def Watch(self, opts):
		"""Watch([]int opts) object, str"""
		return WrappedWatchInterface(handle=_wrapper.client_WrappedNamespaceableResourceInterface_Watch(self.handle, opts.handle))

# Python type for interface client.WrappedResourceInterface
class WrappedResourceInterface(go.GoClass):
	"""The WrappedResourceInterface is designed to mirror client-go's ResourceInterface with one major\ndifference. Instead of accepting and returning pointers to the various apimachinery types/objects\nit handles everything though binary buffers (*uint16) pointers which greatly simplifies the CGO\nfunction call\n"""
	def __init__(self, *args, **kwargs):
		"""
		handle=A Go-side object is always initialized with an explicit handle=arg
		"""
		if len(kwargs) == 1 and 'handle' in kwargs:
			self.handle = kwargs['handle']
			_wrapper.IncRef(self.handle)
		elif len(args) == 1 and isinstance(args[0], go.GoClass):
			self.handle = args[0].handle
			_wrapper.IncRef(self.handle)
		else:
			self.handle = 0
	def Apply(self, name, obj, opts, *args):
		"""Apply(str name, []int obj, []int opts, []str subresources) []int, str"""
		subresources = go.Slice_string(args)
		return go.Slice_byte(handle=_wrapper.client_WrappedResourceInterface_Apply(self.handle, name, obj.handle, opts.handle, subresources.handle))
	def ApplyStatus(self, name, obj, opts):
		"""ApplyStatus(str name, []int obj, []int opts) []int, str"""
		return go.Slice_byte(handle=_wrapper.client_WrappedResourceInterface_ApplyStatus(self.handle, name, obj.handle, opts.handle))
	def Create(self, obj, opts, *args):
		"""Create([]int obj, []int opts, []str subresources) []int, str"""
		subresources = go.Slice_string(args)
		return go.Slice_byte(handle=_wrapper.client_WrappedResourceInterface_Create(self.handle, obj.handle, opts.handle, subresources.handle))
	def Delete(self, name, opts, *args):
		"""Delete(str name, []int opts, []str subresources) str"""
		subresources = go.Slice_string(args)
		return _wrapper.client_WrappedResourceInterface_Delete(self.handle, name, opts.handle, subresources.handle)
	def DeleteCollection(self, opts, listOptions):
		"""DeleteCollection([]int opts, []int listOptions) str"""
		return _wrapper.client_WrappedResourceInterface_DeleteCollection(self.handle, opts.handle, listOptions.handle)
	def Get(self, name, opts, *args):
		"""Get(str name, []int opts, []str subresources) []int, str"""
		subresources = go.Slice_string(args)
		return go.Slice_byte(handle=_wrapper.client_WrappedResourceInterface_Get(self.handle, name, opts.handle, subresources.handle))
	def List(self, opts):
		"""List([]int opts) []int, str"""
		return go.Slice_byte(handle=_wrapper.client_WrappedResourceInterface_List(self.handle, opts.handle))
	def Patch(self, name, pt, data, opts, *args):
		"""Patch(str name, str pt, []int data, []int opts, []str subresources) []int, str"""
		subresources = go.Slice_string(args)
		return go.Slice_byte(handle=_wrapper.client_WrappedResourceInterface_Patch(self.handle, name, pt, data.handle, opts.handle, subresources.handle))
	def Update(self, obj, opts, *args):
		"""Update([]int obj, []int opts, []str subresources) []int, str"""
		subresources = go.Slice_string(args)
		return go.Slice_byte(handle=_wrapper.client_WrappedResourceInterface_Update(self.handle, obj.handle, opts.handle, subresources.handle))
	def UpdateStatus(self, obj, opts):
		"""UpdateStatus([]int obj, []int opts) []int, str"""
		return go.Slice_byte(handle=_wrapper.client_WrappedResourceInterface_UpdateStatus(self.handle, obj.handle, opts.handle))
	def Watch(self, opts):
		"""Watch([]int opts) object, str"""
		return WrappedWatchInterface(handle=_wrapper.client_WrappedResourceInterface_Watch(self.handle, opts.handle))

# Python type for interface client.WrappedWatchInterface
class WrappedWatchInterface(go.GoClass):
	"""WrappedWatchInterface allows CGO applications to iterate over a channel\n"""
	def __init__(self, *args, **kwargs):
		"""
		handle=A Go-side object is always initialized with an explicit handle=arg
		"""
		if len(kwargs) == 1 and 'handle' in kwargs:
			self.handle = kwargs['handle']
			_wrapper.IncRef(self.handle)
		elif len(args) == 1 and isinstance(args[0], go.GoClass):
			self.handle = args[0].handle
			_wrapper.IncRef(self.handle)
		else:
			self.handle = 0
	def Next(self):
		"""Next() []int, str"""
		return go.Slice_byte(handle=_wrapper.client_WrappedWatchInterface_Next(self.handle))
	def Stop(self, goRun=False):
		"""Stop() """
		_wrapper.client_WrappedWatchInterface_Stop(self.handle, goRun)


# ---- Structs ---


# ---- Slices ---


# ---- Maps ---


# ---- Constructors ---


# ---- Functions ---
def NewWrappedDynamicClient(qps, burst, timeout):
	"""NewWrappedDynamicClient(float qps, int burst, str timeout) object, str"""
	return WrappedDynamicClient(handle=_wrapper.client_NewWrappedDynamicClient(qps, burst, timeout))
def NewWrappedDynamicClientWithConfig(config, qps, burst, timeout):
	"""NewWrappedDynamicClientWithConfig([]int config, float qps, int burst, str timeout) object, str"""
	return WrappedDynamicClient(handle=_wrapper.client_NewWrappedDynamicClientWithConfig(config.handle, qps, burst, timeout))
def NewWrappedWatchInterface(watcher):
	"""NewWrappedWatchInterface(object watcher) object
	
	Construct a new WrappedWatchInterface from a client-go watch interface
	"""
	return WrappedWatchInterface(handle=_wrapper.client_NewWrappedWatchInterface(watcher.handle))





# ***************************************************************** #
# (C) Copyright IBM Corporation 2024.                               #
#                                                                   #
# The source code for this program is not published or otherwise    #
# divested of its trade secrets, irrespective of what has been      #
# deposited with the U.S. Copyright Office.                         #
# ***************************************************************** #
# SPDX-License-Identifier: Apache-2.0
"""This module tests that envtest server works with non kubernetes_lite clients"""

from kubernetes_lite.envtest import EnvTest
from kubernetes_lite.envtest.pytest import create_random_namespace
from tests.utils import (  # noqa: F401
    DEFAULT_OBJ_NAME,
    OBJ_API_VERSION,
    OBJ_KIND,
    check_object_status,
    generate_test_object,
    validate_result,
)

import kubernetes
from kubernetes.dynamic import DynamicClient


def test_envtest(kubernetes_env: EnvTest):
    """Test that envtest can work with nonkubernetes_lite libraries"""
    kube_config = kubernetes.config.new_client_from_config(config_file=kubernetes_env.config())
    kube_client = DynamicClient(kube_config)
    assert kube_client

    namespace = create_random_namespace(kubernetes_env.client())

    deploy_res = kube_client.resources.get(kind=OBJ_KIND, api_version=OBJ_API_VERSION)

    obj = generate_test_object(namespace)
    deploy_res.server_side_apply(
        obj,
        name=DEFAULT_OBJ_NAME,
        namespace=namespace,
        field_manager="tet",
    )
    check_object_status(kubernetes_env.client().resource(OBJ_API_VERSION, OBJ_KIND), expected_obj=obj)



# ***************************************************************** #
# (C) Copyright IBM Corporation 2024.                               #
#                                                                   #
# The source code for this program is not published or otherwise    #
# divested of its trade secrets, irrespective of what has been      #
# deposited with the U.S. Copyright Office.                         #
# ***************************************************************** #
# SPDX-License-Identifier: Apache-2.0
"""This module contains common helpers used throughout the various
python scripts
"""

import platform
from enum import StrEnum


# Consts for platforms
class SystemTypes(StrEnum):
    """SystemTypes is a simple enum to track the parent OS"""

    LINUX = "linux"
    DARWIN = "darwin"
    WINDOWS = "windows"


class ArchTypes(StrEnum):
    """ArchTypes is a simple enum to track the processor type"""

    AMD64 = "amd64"
    X386 = "386"
    ARM64 = "arm64"
    S390 = "s390x"


# Parse the system architecture and platform
SYSTEM: SystemTypes | None = None
ARCH: ArchTypes | None = None
if platform.machine().lower() in {"x86_64", "amd64"}:
    ARCH = ArchTypes.AMD64
elif platform.machine().lower() in {"i386", "i686"}:
    ARCH = ArchTypes.X386
elif platform.machine().lower() in {"aarch64", "arm64", "armv8b", "armv8l", "aarch64_be"}:
    ARCH = ArchTypes.ARM64
elif platform.machine().lower() in {"s390x"}:
    ARCH = ArchTypes.S390
else:
    raise ValueError(f"Unknown machine platform {platform.machine()}")

if platform.system() == "Linux":
    SYSTEM = SystemTypes.LINUX
elif platform.system() == "Darwin":
    SYSTEM = SystemTypes.DARWIN
elif platform.system() == "Windows":
    SYSTEM = SystemTypes.WINDOWS
else:
    raise ValueError(f"Unknown machine system {platform.system()}")



"""
 ----------------------------------------------------------------- *
 * (C) Copyright IBM Corporation 2024.                               *
 *                                                                   *
 * The source code for this program is not published or otherwise    *
 * divested of its trade secrets, irrespective of what has been      *
 * deposited with the U.S. Copyright Office.                         *
 * -----------------------------------------------------------------

 ----------------------------------------------------------------- *
 * (C) Copyright IBM Corporation 2024.                               *
 *                                                                   *
 * The source code for this program is not published or otherwise    *
 * divested of its trade secrets, irrespective of what has been      *
 * deposited with the U.S. Copyright Office.                         *
 * -----------------------------------------------------------------

"""
# python wrapper for package github.ibm.com/Michael-Honaker/kubernetes-lite/kubernetes_lite/go_wrapper/pkg/envtest/setup within overall package wrapper
# This is what you import to use the package.
# File is generated by gopy. Do not edit.
# gopy pkg -no-make -author=Michael Honaker -email=michael.honaker@ibm.com -name=wrapper github.ibm.com/Michael-Honaker/kubernetes-lite/kubernetes_lite/go_wrapper/pkg/client github.ibm.com/Michael-Honaker/kubernetes-lite/kubernetes_lite/go_wrapper/pkg/envtest/setup github.ibm.com/Michael-Honaker/kubernetes-lite/kubernetes_lite/go_wrapper/pkg/envtest/server

# the following is required to enable dlopen to open the _go.so file
import os,sys,inspect,collections
try:
	import collections.abc as _collections_abc
except ImportError:
	_collections_abc = collections

cwd = os.getcwd()
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
os.chdir(currentdir)
from . import _wrapper
from . import go

os.chdir(cwd)

# to use this code in your end-user python file, import it as follows:
# from wrapper import setup
# and then refer to everything using setup. prefix
# packages imported by this package listed below:




# ---- Types ---


#---- Enums from Go (collections of consts with same type) ---


#---- Constants from Go: Python can only ask that you please don't change these! ---


# ---- Global Variables: can only use functions to access ---


# ---- Interfaces ---


# ---- Structs ---

# Python type for struct setup.SetupEnvTestResult
class SetupEnvTestResult(go.GoClass):
	"""SetupEnvTestResult contains the various results from running the SetupEnvTest command\nincludes stdout/stderr and the serialized error message\n"""
	def __init__(self, *args, **kwargs):
		"""
		handle=A Go-side object is always initialized with an explicit handle=arg
		otherwise parameters can be unnamed in order of field names or named fields
		in which case a new Go object is constructed first
		"""
		if len(kwargs) == 1 and 'handle' in kwargs:
			self.handle = kwargs['handle']
			_wrapper.IncRef(self.handle)
		elif len(args) == 1 and isinstance(args[0], go.GoClass):
			self.handle = args[0].handle
			_wrapper.IncRef(self.handle)
		else:
			self.handle = _wrapper.setup_SetupEnvTestResult_CTor()
			_wrapper.IncRef(self.handle)
	def __del__(self):
		_wrapper.DecRef(self.handle)
	def __str__(self):
		pr = [(p, getattr(self, p)) for p in dir(self) if not p.startswith('__')]
		sv = 'setup.SetupEnvTestResult{'
		first = True
		for v in pr:
			if callable(v[1]):
				continue
			if first:
				first = False
			else:
				sv += ', '
			sv += v[0] + '=' + str(v[1])
		return sv + '}'
	def __repr__(self):
		pr = [(p, getattr(self, p)) for p in dir(self) if not p.startswith('__')]
		sv = 'setup.SetupEnvTestResult ( '
		for v in pr:
			if not callable(v[1]):
				sv += v[0] + '=' + str(v[1]) + ', '
		return sv + ')'


# ---- Slices ---


# ---- Maps ---


# ---- Constructors ---


# ---- Functions ---
def SetupEnvTest(stdin_data, arg_data):
	"""SetupEnvTest([]int stdin_data, []int arg_data) []int, str
	
	SetupEnvTest runs the vendored main file with the provided stdin_data and json arg data
	"""
	return go.Slice_byte(handle=_wrapper.setup_SetupEnvTest(stdin_data.handle, arg_data.handle))





"""
 ----------------------------------------------------------------- *
 * (C) Copyright IBM Corporation 2024.                               *
 *                                                                   *
 * The source code for this program is not published or otherwise    *
 * divested of its trade secrets, irrespective of what has been      *
 * deposited with the U.S. Copyright Office.                         *
 * -----------------------------------------------------------------

"""
# python wrapper for package github.ibm.com/Michael-Honaker/kubernetes-lite/kubernetes_lite/go_wrapper/pkg/envtest/server within overall package wrapper
# This is what you import to use the package.
# File is generated by gopy. Do not edit.
# gopy pkg -no-make -author=Michael Honaker -email=michael.honaker@ibm.com -name=wrapper github.ibm.com/Michael-Honaker/kubernetes-lite/kubernetes_lite/go_wrapper/pkg/client github.ibm.com/Michael-Honaker/kubernetes-lite/kubernetes_lite/go_wrapper/pkg/envtest/setup github.ibm.com/Michael-Honaker/kubernetes-lite/kubernetes_lite/go_wrapper/pkg/envtest/server

# the following is required to enable dlopen to open the _go.so file
import os,sys,inspect,collections
try:
	import collections.abc as _collections_abc
except ImportError:
	_collections_abc = collections

cwd = os.getcwd()
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
os.chdir(currentdir)
from . import _wrapper
from . import go

os.chdir(cwd)

# to use this code in your end-user python file, import it as follows:
# from wrapper import server
# and then refer to everything using server. prefix
# packages imported by this package listed below:




# ---- Types ---


#---- Enums from Go (collections of consts with same type) ---


#---- Constants from Go: Python can only ask that you please don't change these! ---
DefaultContextName = "default-context"
"""
Default settings for the kube config

"""
DefaultNamespace = "default"
"""
Default settings for the kube config

"""


# ---- Global Variables: can only use functions to access ---


# ---- Interfaces ---

# Python type for interface server.EnvTestEnvironment
class EnvTestEnvironment(go.GoClass):
	"""EnvTestEnvironment defines the interface for starting and stopping an EnvTest server as well as\nfetching a generic kube config\n"""
	def __init__(self, *args, **kwargs):
		"""
		handle=A Go-side object is always initialized with an explicit handle=arg
		"""
		if len(kwargs) == 1 and 'handle' in kwargs:
			self.handle = kwargs['handle']
			_wrapper.IncRef(self.handle)
		elif len(args) == 1 and isinstance(args[0], go.GoClass):
			self.handle = args[0].handle
			_wrapper.IncRef(self.handle)
		else:
			self.handle = 0
	def GetKubeConfig(self):
		"""GetKubeConfig() []int, str"""
		return go.Slice_byte(handle=_wrapper.server_EnvTestEnvironment_GetKubeConfig(self.handle))
	def Start(self):
		"""Start() []int, str"""
		return go.Slice_byte(handle=_wrapper.server_EnvTestEnvironment_Start(self.handle))
	def Stop(self):
		"""Stop() str"""
		return _wrapper.server_EnvTestEnvironment_Stop(self.handle)


# ---- Structs ---


# ---- Slices ---


# ---- Maps ---


# ---- Constructors ---


# ---- Functions ---
def NewEnvTestEnvironment():
	"""NewEnvTestEnvironment() object, str"""
	return EnvTestEnvironment(handle=_wrapper.server_NewEnvTestEnvironment())
def NewEnvTestEnvironmentWithPath(path):
	"""NewEnvTestEnvironmentWithPath(str path) object, str"""
	return EnvTestEnvironment(handle=_wrapper.server_NewEnvTestEnvironmentWithPath(path))





# ***************************************************************** #
# (C) Copyright IBM Corporation 2024.                               #
#                                                                   #
# The source code for this program is not published or otherwise    #
# divested of its trade secrets, irrespective of what has been      #
# deposited with the U.S. Copyright Office.                         #
# ***************************************************************** #
# SPDX-License-Identifier: Apache-2.0
"""This module contains a dataclass used in processing the result from
setup-envtest
"""

from dataclasses import dataclass
from functools import cached_property
from logging import Logger

from kubernetes_lite.serialization import deserialize_from_go
from kubernetes_lite.setup_envtest.exit_data import ExitData
from kubernetes_lite.wrapper.go import Slice_byte


@dataclass
class SetupEnvTestResult:
    """SetupEnvTestResult contains information regarding the result of a setup-envtest
    run. This includes the captured stdout/stderr as well as the exit information

    Attributes:
        stderr (str | None): the captured stderr data if some was captured
        stdout (str | None): the captured stdout data if some was captured
        error (str | None): the serialized error data if one was returned
    """

    stderr: str | None = None
    stdout: str | None = None
    error: str | None = None

    @cached_property
    def exit_data(self) -> ExitData | None:
        """The parsed ExitData instance for this result

        Returns:
            ExitData | None: The ExitData object if one could be generated. Else None
        """
        if not self.error:
            return
        return ExitData.from_string(self.error)

    @cached_property
    def is_multi_output(self) -> bool:
        """Helper property used when printing the result to determine if
        multiple ==<>== header tags are needed

        Returns:
            bool: If more then one of stdout/stderr/error was provided
        """
        return (
            sum(
                [
                    bool(self.stdout),
                    bool(self.stderr),
                    bool(self.exit_data.message if self.exit_data else False),
                ]
            )
            > 1
        )

    def log_and_raise(self, log: Logger, raise_exc: bool = False):
        """Helper function to print the results to a log and raise an exception
        if the command was not successful

        Args:
            log (Logger): The logger to print result information to
            raise_exc (bool, optional): If an exception should be raised if the result was
                invalid. Defaults to False.
        """
        should_raise_exc: bool = False
        if self.exit_data and self.exit_data.message:
            if self.is_multi_output:
                log.error("==Error==")
            log.error(self.exit_data.message)
            should_raise_exc = True
        if self.stderr:
            if self.is_multi_output:
                log.info("==StdErr==")
            log.error(self.stderr)
            should_raise_exc = True
        if self.stdout:
            if self.is_multi_output:
                log.info("==StdOut==")
            log.info(self.stdout)

        if should_raise_exc and raise_exc:
            raise RuntimeError("Failed to run SetupEvTet")

    @classmethod
    def from_go(cls, raw_data: Slice_byte) -> "SetupEnvTestResult":
        """Parse a SetupEnvTestResult object from a raw golang Slice_byte object

        Args:
            raw_data (Slice_byte): The raw data from go

        Returns:
            SetupEnvTestResult: the parsed result object
        """
        result_dict = deserialize_from_go(raw_data)

        return cls(**result_dict)



# ***************************************************************** #
# (C) Copyright IBM Corporation 2024.                               #
#                                                                   #
# The source code for this program is not published or otherwise    #
# divested of its trade secrets, irrespective of what has been      #
# deposited with the U.S. Copyright Office.                         #
# ***************************************************************** #
# SPDX-License-Identifier: Apache-2.0
"""This module contains code around sanitizing an object for deployment"""

import datetime
from dataclasses import dataclass
from typing import Any

from kubernetes_lite.serialization import serialize_to_go
from kubernetes_lite.wrapper.go import Slice_byte

# SourceObjectType are helper types to attempt to describe common objects passed to client functions. See
# sanitize_resource_for_serialization for more information
SourceObjectType = (
    float | bool | property | datetime.datetime | datetime.date | bytes | str | int | list | tuple | dict | Any | None
)
SanitizedObjectType = float | bool | bytes | str | int | list | tuple | dict | Any | None


# Stolen from kubernetes but modified for None pruning, safety, and openapi handling
# https://github.com/kubernetes-client/python/blob/d67bc8c2bdb89b29c17c1ba0edb03a48d977c0e2/kubernetes/client/api_client.py#L202
def sanitize_resource_for_serialization(obj: SourceObjectType) -> SanitizedObjectType:  # noqa: PLR0911, PLR0912
    """Sanitize an object before serializing it to bytes
    If obj is None, return None.
    If obj is str, int, long, float, bool, return directly.
    If obj is datetime.datetime, datetime.date
        convert to string in iso8601 format.
    If obj is list, sanitize each element in the list.
    If obj is dict, return the dict.
    If obj is OpenAPI model, return the properties dict.
    :param obj: The data to serialize.
    :return: The serialized form of data.
    """
    if obj is None:  # pylint: disable=no-else-return
        return None
    elif isinstance(obj, float | bool | bytes | str | int):
        return obj
    elif isinstance(obj, list):
        return [sanitize_resource_for_serialization(sub_obj) for sub_obj in obj]
    elif isinstance(obj, tuple):
        return tuple(sanitize_resource_for_serialization(sub_obj) for sub_obj in obj)
    elif isinstance(obj, datetime.datetime | datetime.date):
        return obj.isoformat()
    elif isinstance(obj, property):
        return sanitize_resource_for_serialization(obj.fget())

    if isinstance(obj, dict):
        obj_dict = obj
    elif hasattr(obj, "attribute_map"):
        # Convert model obj to dict except
        # `openapi_types` and `attribute_map`.
        # Convert attribute name to json key in
        # model definition for request.
        obj_dict = {}
        for attr, name in obj.attribute_map.items():
            if hasattr(obj, attr):
                obj_dict[name] = getattr(obj, attr)
    else:
        # Hope that the object can still be serialized by the json encoder.
        return obj

    # Prune fields which are None but keep
    # empty arrays or dictionaries
    return_dict = {}
    for key, val in obj_dict.items():
        updated_obj = sanitize_resource_for_serialization(val)
        if updated_obj is not None:
            return_dict[key] = updated_obj
    return return_dict


@dataclass
class ObjectMetadata:
    """ObjectMetadata is used to describe extracted information from an object

    Attributes:
        metadata (dict[str, Any]): Metadata extracted from the object
        name (str):  Name of the object
        namespace (str | None): Optional namespace of the object

    """

    metadata: dict[str, Any]
    name: str
    namespace: str | None


def process_input_resource(obj: SanitizedObjectType) -> tuple[Slice_byte, ObjectMetadata]:
    """Sanitize and serialize an object into go. After sanitizing extract some metadata about the
    object.

    Args:
        obj (SanitizedObjectType): The object to sanitize/serialize

    Returns:
        tuple[Slice_byte, ObjectMetadata]: The serialized go object and extracted metadata
    """
    sanitized_obj = sanitize_resource_for_serialization(obj)

    if not isinstance(sanitized_obj, dict):
        raise RuntimeError(f"can not deploy object of type {type(sanitized_obj)}")

    # Extract metadata from sanitized object
    metadata = ObjectMetadata(
        name=sanitized_obj.get("metadata", {}).get("name"),
        namespace=sanitized_obj.get("metadata", {}).get("namespace"),
        metadata=sanitized_obj.get("metadata", {}),
    )

    return (serialize_to_go(sanitized_obj), metadata)



# ***************************************************************** #
# (C) Copyright IBM Corporation 2024.                               #
#                                                                   #
# The source code for this program is not published or otherwise    #
# divested of its trade secrets, irrespective of what has been      #
# deposited with the U.S. Copyright Office.                         #
# ***************************************************************** #
# SPDX-License-Identifier: Apache-2.0
"""This module contains functions to run setup-envtest commands"""

import copy
import logging
import select
import sys
from logging import Logger
from pathlib import Path

from kubernetes_lite.serialization import serialize_to_go
from kubernetes_lite.setup_envtest.result import SetupEnvTestResult
from kubernetes_lite.wrapper.setup import SetupEnvTest


def internal_run_setup_envtest_command(*args: list[str], stdin: bytes = b"") -> SetupEnvTestResult:
    """This function runs the underlying setup_envtest command with a provided args and stdin.

    Args:
        *args (list[str]): The arguments to pass to setup-envtest
        stdin (bytes, optional): The stdin data to send to setup-envtest. Defaults to b"".

    Returns:
        SetupEnvTestResult: A parsed dataclass containing the returned error and captured stdout/stderr
    """
    # Convert input args to go slices
    go_stdin = serialize_to_go(stdin)
    go_args = serialize_to_go(["python3 -m kubernetes_lite.setup_envtest", *args])

    # Run setup env test
    setup_result_raw = SetupEnvTest(go_stdin, go_args)

    # Deserialize the result
    return SetupEnvTestResult.from_go(setup_result_raw)


def run_setup_envtest_command(log: Logger | None = None):
    """Function to handle running the setup_envtest command as if the user ran this module
    directly. It handles reading from the process args/stdin and passing them to
    internal_run_setup_envtest_command. This function handles logging the result and
    exiting with the correct exit code

    Args:
        log (Logger | None, optional): Logger to use for printing setup_envtest messages.
            Defaults to None.
    """
    if not log:
        log = logging.getLogger("setup-envtest")

    args = copy.deepcopy(sys.argv)

    # If there are args in sys.args then check to see if the module was included. If so remove the first item
    if len(args) > 0:
        current_file_path = Path(__file__)
        if str(current_file_path.parent) in args[0]:
            args.pop(0)

    # Read stdin into a buffer
    stdin_available_list, _, _ = select.select([sys.stdin], [], [], 0.1)
    stdin_data = ""
    if len(stdin_available_list) > 0:
        stdin_byte = sys.stdin.read(1)
        stdin_data = stdin_byte
        while len(stdin_byte) > 0:
            stdin_byte = sys.stdin.read(1)
            stdin_data += stdin_byte

    setup_result = internal_run_setup_envtest_command(*args, stdin=stdin_data.encode("utf-8"))

    # Log results messages without raising the exception
    setup_result.log_and_raise(log)

    # If we parsed and exit code from the logs then exit
    if setup_result.exit_data and setup_result.exit_data.code:
        sys.exit(setup_result.exit_data.code)



# ***************************************************************** #
# (C) Copyright IBM Corporation 2024.                               #
#                                                                   #
# The source code for this program is not published or otherwise    #
# divested of its trade secrets, irrespective of what has been      #
# deposited with the U.S. Copyright Office.                         #
# ***************************************************************** #
# SPDX-License-Identifier: Apache-2.0
"""This module contains the abstraction and implementation of various kubernetes
clients to aid in performance testing
"""

from performance.interfaces.base import ClientInterface  # noqa: F401



# ***************************************************************** #
# (C) Copyright IBM Corporation 2024.                               #
#                                                                   #
# The source code for this program is not published or otherwise    #
# divested of its trade secrets, irrespective of what has been      #
# deposited with the U.S. Copyright Office.                         #
# ***************************************************************** #
# SPDX-License-Identifier: Apache-2.0
"""This module contains helper pytest fixtures to make using EnvTest easier
in testing environments
"""

import random
import string
from collections.abc import Generator
from contextlib import contextmanager

from kubernetes_lite.client import DynamicClient
from kubernetes_lite.envtest import EnvTest

import pytest


@contextmanager
def yield_kubernetes_env() -> Generator[EnvTest, None, None]:
    """Context manager to yield a running EnvTest instance and handle
    stopping it after the fact

    Yields:
        Generator[EnvTest, None, None]: The running EnvTest instance
    """
    env = EnvTest()
    env.start()
    yield env
    env.stop()


@pytest.fixture()
def kubernetes_env() -> Generator[EnvTest, None, None]:
    """Pytest fixture for a running EnvTest instance

    Yields:
        Generator[EnvTest, None, None]: The running EnvTest instance
    """
    with yield_kubernetes_env() as env:
        yield env


@pytest.fixture(scope="session")
def session_kubernetes_env() -> Generator[EnvTest, None, None]:
    """Session scoped version of the above fixture. Returns a running
    EnvTest instance

    Yields:
        Generator[EnvTest, None, None]: The running EnvTest instance
    """
    with yield_kubernetes_env() as env:
        yield env


@pytest.fixture()
def dynamic_client(kubernetes_env: EnvTest) -> Generator[DynamicClient, None, None]:
    """A pytest fixture for a dynamic client connected to an EnvTest instance

    Args:
        kubernetes_env (EnvTest): a running EnvTest instance

    Yields:
        Generator[DynamicClient, None, None]: The dynamic client that can be used to interact
            with an EnvTest instance
    """
    yield kubernetes_env.client()


@pytest.fixture(scope="session")
def session_client(session_kubernetes_env: EnvTest) -> Generator[DynamicClient, None, None]:
    """A session scoped pytest fixture for a dynamic client connected to an EnvTest
    instance

    Args:
        session_kubernetes_env (EnvTest): a running EnvTest instance

    Yields:
        Generator[DynamicClient, None, None]: The dynamic client that can be used to interact
            with an EnvTest instance
    """
    yield session_kubernetes_env.client()


def create_random_namespace(client: DynamicClient) -> str:
    """Helper function to create a randomly named namespace from the given client

    Args:
        client (DynamicClient): The client to create namespaces with

    Returns:
        str: The randomly generated namespace name
    """
    # Randomize the namespace string
    namespace = "".join(random.choice(string.ascii_lowercase) for i in range(10))
    client.resources.get(api_version="v1", kind="Namespace").apply(
        {"metadata": {"name": namespace}, "kind": "Namespace", "apiVersion": "v1"}, fieldManager="default"
    )
    return namespace


@pytest.fixture()
def dynamic_client_namespace(dynamic_client: DynamicClient) -> Generator[tuple[DynamicClient, str], None, None]:
    """Pytest fixture to return a Dynamic Client with a fresh namespace you can apply resources to

    Args:
        dynamic_client (DynamicClient): Dynamic Client fixture

    Yields:
        Generator[tuple[DynamicClient, str], None, None]: tuple with the client and namespace name
    """
    yield dynamic_client, create_random_namespace(dynamic_client)


@pytest.fixture()
def session_client_namespace(session_client: DynamicClient) -> Generator[tuple[DynamicClient, str], None, None]:
    """Pytest session scoped fixture to return a Dynamic Client with a fresh namespace you can apply resources to

    Args:
        session_client (DynamicClient): Dynamic Client fixture

    Yields:
        Generator[tuple[DynamicClient, str], None, None]: tuple with the client and namespace name
    """
    yield session_client, create_random_namespace(session_client)



# ***************************************************************** #
# (C) Copyright IBM Corporation 2024.                               #
#                                                                   #
# The source code for this program is not published or otherwise    #
# divested of its trade secrets, irrespective of what has been      #
# deposited with the U.S. Copyright Office.                         #
# ***************************************************************** #
# SPDX-License-Identifier: Apache-2.0
"""This module contains the helper functions/threads for tracking memory
usage
"""

import ctypes
import gc
import sys
import time
from threading import Event, Lock, Thread

import psutil


def get_current_memory_usage() -> int:
    """Get the current RSS memory usage for this process

    Returns:
        int: The current memory usage in bytes
    """
    current_process = psutil.Process()
    gc.collect()
    return current_process.memory_info().rss


class BackgroundMemoryTracker(Thread):
    """BackgroundMemoryTracker is a background thread that constantly records the
    maximum memory used since starting/being reset
    """

    def __init__(self, *args, poll_time: float = 0.1, **kwargs):
        """Initialize the BackgroundMemoryTracker and all required synchronization primitives

        Args:
            poll_time (float, optional): How often to check for max memory. Defaults to 0.1.
            *args: Extra args passed to Thread __init__
            **kwargs: Extra kwargs pass to Thread __init__
        """
        super().__init__(*args, **kwargs, daemon=True)

        self.poll_time = poll_time

        self.shutdown_event = Event()
        self.memory_lock = Lock()
        self.max_mem = get_current_memory_usage()

    def run(self):
        """Constantly poll the system for max memory until the shutdown event is set"""
        while True:
            if self.shutdown_event.is_set():
                return

            with self.memory_lock:
                current_usage = get_current_memory_usage()
                self.max_mem = max(current_usage, self.max_mem)
            time.sleep(self.poll_time)

    def get_max_mem(self) -> int:
        """Get the max memory observed by this thread

        Returns:
            int: max memory since last reset/start
        """
        return self.max_mem

    def reset(self):
        """Reset the memory tracker back to the current usage"""
        with self.memory_lock:
            gc.collect()
            if sys.platform.startswith("linux"):
                libc = ctypes.CDLL("libc.so.6")
                libc.malloc_trim(0)

            self.max_mem = get_current_memory_usage()

    def shutdown(self):
        """Shutdown the background thread by setting an event flag"""
        self.shutdown_event.set()



# ***************************************************************** #
# (C) Copyright IBM Corporation 2024.                               #
#                                                                   #
# The source code for this program is not published or otherwise    #
# divested of its trade secrets, irrespective of what has been      #
# deposited with the U.S. Copyright Office.                         #
# ***************************************************************** #
# SPDX-License-Identifier: Apache-2.0
"""This module contains functions to run setup-envtest. See
https://pkg.go.dev/sigs.k8s.io/controller-runtime/tools/setup-envtest for
more information regarding arguments
"""

from kubernetes_lite.setup_envtest.command import (
    internal_run_setup_envtest_command,
    run_setup_envtest_command,
)

__all__ = ["internal_run_setup_envtest_command", "run_setup_envtest_command"]



from __future__ import annotations

import platform
import sys
import os

from setuptools import Extension
from setuptools import setup

if sys.platform != 'win32' and platform.python_implementation() == 'CPython':
    try:
        import wheel.bdist_wheel
    except ImportError:
        cmdclass = {}
    else:
        class bdist_wheel(wheel.bdist_wheel.bdist_wheel):
            def finalize_options(self) -> None:
                self.py_limited_api = f'cp3{sys.version_info[1]}'
                super().finalize_options()

        cmdclass = {'bdist_wheel': bdist_wheel}
else:
    cmdclass = {}


os.environ["CGO_CFLAGS_ALLOW"]="-O.*"

setup(
    ext_modules=[
        Extension(
            'kubernetes_lite.wrapper._wrapper', ['kubernetes_lite/wrapper/wrapper.go'],
            py_limited_api=True,
        ),
    ],
    cmdclass=cmdclass,
    build_golang={'root': 'github.ibm.com/Michael-Honaker/kubernetes-lite/kubernetes_lite'},
)


# ***************************************************************** #
# (C) Copyright IBM Corporation 2024.                               #
#                                                                   #
# The source code for this program is not published or otherwise    #
# divested of its trade secrets, irrespective of what has been      #
# deposited with the U.S. Copyright Office.                         #
# ***************************************************************** #
# SPDX-License-Identifier: Apache-2.0
"""This module runs the setup-envtest command from python as if the user ran
setuptest-env directly
"""

import logging

from kubernetes_lite.setup_envtest import run_setup_envtest_command

# Configure logging and run the setup_envtest command
logging.basicConfig(level=logging.INFO, format="%(message)s")
run_setup_envtest_command()



# ***************************************************************** #
# (C) Copyright IBM Corporation 2024.                               #
#                                                                   #
# The source code for this program is not published or otherwise    #
# divested of its trade secrets, irrespective of what has been      #
# deposited with the U.S. Copyright Office.                         #
# ***************************************************************** #
# SPDX-License-Identifier: Apache-2.0
"""This module contains the helper functions used to test an interface"""

import json
import time
from contextlib import suppress
from dataclasses import dataclass, field
from pathlib import Path

from performance.interfaces import ClientInterface
from performance.utils import time_for_list, unimport_module

ROOT_FOLDER_PATH = Path(__file__).parent.parent


@dataclass
class ResourceTimingInfo:
    """ResourceTimingInfo contains the time information for functions that change based
    on the size of the input

    Attributes:
        create_times (list[float]): List of times required to create an object
        delete_times (list[float]): List of times required to delete an object
        apply_times (list[float]): List of times required to apply an object
        list_times (list[float]): List of times required to list objects
        get_times (list[float]): List of times required to get an object
    """

    create_times: list[float] = field(default_factory=list)
    delete_times: list[float] = field(default_factory=list)
    apply_times: list[float] = field(default_factory=list)
    list_times: list[float] = field(default_factory=list)
    get_times: list[float] = field(default_factory=list)


@dataclass
class TimingInfo:
    """TimingInfo contains the time information for constant fields like
    import/client_creation

    Attributes:
        import_times (list[float]): List of times for importing the module
        client_create_times (list[float]): List of times for creating the client
        resource_client_create_times (list[float]): List of times for creating the resource client
        resource_timing_info (dict[str, ResourceTimingInfo]): Mapping of timing results to the resource name
    """

    import_times: list[float] = field(default_factory=list)
    client_create_times: list[float] = field(default_factory=list)
    resource_client_create_times: list[float] = field(default_factory=list)
    resource_timing_info: dict[str, ResourceTimingInfo] = field(default_factory=dict)


def test_interface(client_interface: ClientInterface, repetitions: int = 5, namespace: str = "default") -> TimingInfo:
    """Test Interface is a helper function used to test an individual interface across multiple
    objects

    Args:
        client_interface (ClientInterface): The client interface being tested
        repetitions (int, optional): How many repetitions to do. Defaults to 5.
        namespace (str, optional): Namespace to run tests in. Defaults to "default".

    Returns:
        TimingInfo: The timing information for this client
    """
    timing_info = TimingInfo()

    print(f"{client_interface.__class__.__name__}: Testing import")

    # Time how long import takes
    modules_to_unimport = []
    for _ in range(repetitions):
        with time_for_list(timing_info.import_times):
            modules_to_unimport = client_interface.import_modules()
        for module in modules_to_unimport:
            unimport_module(module)

    # Redo the import outside of timing so the rest can use it
    client_interface.import_modules()

    print(f"{client_interface.__class__.__name__}: Testing client creation")

    # Time how long client creation takes
    for _ in range(repetitions):
        with time_for_list(timing_info.client_create_times):
            client = client_interface.create_client()

    # Setup the data and thing to test
    test_crd = json.loads((ROOT_FOLDER_PATH / Path("test_crd.json")).read_bytes())
    test_crd_kind = test_crd.get("kind")
    test_crd_api_version = test_crd.get("apiVersion")
    crd_resource_client = client_interface.create_resource_client(client, test_crd_kind, test_crd_api_version)
    client_interface.apply(crd_resource_client, test_crd)

    # Wait for CRD to take effect
    time.sleep(1)

    # Time how long client resource creation takes
    for _ in range(repetitions):
        with time_for_list(timing_info.resource_client_create_times):
            resource_client = client_interface.create_resource_client(client, test_crd_kind, test_crd_api_version)

    # For each file in the data folder try running the test
    for test_file in ROOT_FOLDER_PATH.glob("data/*.json"):
        print(f"{client_interface.__class__.__name__}: Testing file: {test_file}")
        data_to_test = json.loads(test_file.read_bytes())
        data_to_test_name = f"{data_to_test['metadata']['name']}-{client_interface.__class__.__name__}".lower()
        data_to_test_namespace = namespace
        data_to_test_kind = data_to_test.get("kind")
        data_to_test_api_version = data_to_test.get("apiVersion")

        # Update CR with unique names/namespace
        data_to_test["metadata"]["namespace"] = data_to_test_namespace
        data_to_test["metadata"]["name"] = data_to_test_name
        resource_client = client_interface.create_resource_client(client, data_to_test_kind, data_to_test_api_version)

        # Delete any existing resource incase it was accidentally created
        with suppress(Exception):
            client_interface.delete(
                resource_client, data_to_test_name, data_to_test_namespace, data_to_test_kind, data_to_test_api_version
            )

        # For the number of repetitions run the test
        resource_timing_info = ResourceTimingInfo()
        for _ in range(repetitions):
            with time_for_list(resource_timing_info.create_times):
                client_interface.create(resource_client, data_to_test)
            with time_for_list(resource_timing_info.apply_times):
                client_interface.apply(resource_client, data_to_test)
            with time_for_list(resource_timing_info.get_times):
                client_interface.get(
                    resource_client,
                    data_to_test_name,
                    data_to_test_namespace,
                )
            with time_for_list(resource_timing_info.list_times):
                client_interface.list(resource_client, data_to_test_namespace)
            with time_for_list(resource_timing_info.delete_times):
                client_interface.delete(
                    resource_client,
                    data_to_test_name,
                    data_to_test_namespace,
                )
        timing_info.resource_timing_info[test_file.name] = resource_timing_info

    # Unimport the modules for better memory tracking
    for module in modules_to_unimport:
        unimport_module(module)

    return timing_info
